# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy
import os
import re
import urllib.error
import urllib.request
from typing import TYPE_CHECKING, Any, TypedDict

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

from streamlit import cli_util, url_util
from streamlit.config_option import ConfigOption
from streamlit.elements.lib.color_util import is_css_color_like
from streamlit.errors import (
    StreamlitInvalidThemeError,
    StreamlitInvalidThemeOptionError,
    StreamlitInvalidThemeSectionError,
)

# Maximum size for theme files (1MB). Theme files should be small configuration
# files containing only theme options, not large data files.
_MAX_THEME_FILE_SIZE_BYTES = 1024 * 1024  # 1MB
_CONFIG_FILE_SCOPES = ("global", "project", "script")


class _ThemeOverride(TypedDict):
    value: Any
    where_defined: str


def _format_config_file_label(
    where_defined: str | None, config_file_paths: Sequence[str] | None
) -> str | None:
    """Return a scope-labeled config.toml path if where_defined matches a config file."""
    if not where_defined or not config_file_paths:
        return None

    for index, path in enumerate(config_file_paths):
        if where_defined == path:
            scope = (
                _CONFIG_FILE_SCOPES[index] if index < len(_CONFIG_FILE_SCOPES) else None
            )
            if scope:
                return f"config.toml ({scope}): {path}"
            return f"config.toml: {path}"

    return None


def _get_logger() -> Any:
    """Get logger for this module. Separate function to avoid circular imports."""
    from streamlit.logger import get_logger

    return get_logger(__name__)


def server_option_changed(
    old_options: dict[str, ConfigOption], new_options: dict[str, ConfigOption]
) -> bool:
    """Return True if and only if an option in the server section differs
    between old_options and new_options.
    """
    for opt_name, opt_val in old_options.items():
        if not opt_name.startswith("server"):
            continue

        old_val = opt_val.value
        new_val = new_options[opt_name].value
        if old_val != new_val:
            return True

    return False


def show_config(
    section_descriptions: dict[str, str],
    config_options: dict[str, ConfigOption],
) -> None:
    """Print the given config sections/options to the terminal."""

    out = []
    out.append(
        _clean(
            """
        # Below are all the sections and options you can have in
        ~/.streamlit/config.toml.
    """
        )
    )

    def append_desc(text: str) -> None:
        out.append("# " + cli_util.style_for_cli(text, bold=True))

    def append_comment(text: str) -> None:
        out.append("# " + cli_util.style_for_cli(text))

    def append_section(text: str) -> None:
        out.append(cli_util.style_for_cli(text, bold=True, fg="green"))

    def append_setting(text: str) -> None:
        out.append(cli_util.style_for_cli(text, fg="green"))

    for section in section_descriptions:
        # We inject a fake config section used for unit tests that we exclude here as
        # its options are often missing required properties, which confuses the code
        # below.
        if section == "_test":
            continue

        section_options = {
            k: v
            for k, v in config_options.items()
            if v.section == section and v.visibility == "visible" and not v.is_expired()
        }

        # Only show config header if section is non-empty.
        if len(section_options) == 0:
            continue

        out.append("")
        append_section(f"[{section}]")
        out.append("")

        for option in section_options.values():
            key = option.key.split(".")[-1]
            description_paragraphs = _clean_paragraphs(option.description or "")

            last_paragraph_idx = len(description_paragraphs) - 1

            for i, paragraph in enumerate(description_paragraphs):
                # Split paragraph into lines
                lines = paragraph.rstrip().split(
                    "\n"
                )  # Remove trailing newline characters

                # If the first line is empty, remove it
                if lines and not lines[0].strip():
                    lines = lines[1:]

                # Choose function based on whether it's the first paragraph or not
                append_func = append_desc if i == 0 else append_comment

                # Add comment character to each line and add to out
                for line in lines:
                    append_func(line.lstrip())

                # # Add a line break after a paragraph only if it's not the last paragraph
                if i != last_paragraph_idx:
                    append_comment("")

            if option.deprecated:
                if out[-1] != "#":
                    append_comment("")
                append_comment(
                    cli_util.style_for_cli("THIS IS DEPRECATED.", fg="yellow")
                )
                append_comment("")
                for line in _clean_paragraphs(option.deprecation_text):
                    append_comment(line)
                append_comment("")
                append_comment(
                    f"This option will be removed on or after {option.expiration_date}."
                )

            import toml

            toml_default = toml.dumps({"default": option.default_val})
            toml_default = toml_default[10:].strip()

            if len(toml_default) > 0:
                # Ensure a line break before appending "Default" comment, if not already there
                if out[-1] != "#":
                    append_comment("")
                append_comment(f"Default: {toml_default}")
            else:
                # Don't say "Default: (unset)" here because this branch applies
                # to complex config settings too.
                pass

            option_is_manually_set = (
                option.where_defined != ConfigOption.DEFAULT_DEFINITION
            )

            if option_is_manually_set:
                if out[-1] != "# ":
                    append_comment("")
                append_comment(f"The value below was set in {option.where_defined}")

            toml_setting = toml.dumps({key: option.value})

            if len(toml_setting) == 0:
                toml_setting = f"# {key} =\n"
            elif not option_is_manually_set:
                toml_setting = f"# {toml_setting}"

            append_setting(toml_setting)

    cli_util.print_to_cli("\n".join(out))


def _clean(txt: str) -> str:
    """Replace sequences of multiple spaces with a single space, excluding newlines.

    Preserves leading and trailing spaces, and does not modify spaces in between lines.
    """
    return re.sub(r" +", " ", txt)


def _clean_paragraphs(txt: str) -> list[str]:
    """Split the text into paragraphs, preserve newlines within the paragraphs."""
    # Strip both leading and trailing newlines.
    txt = txt.strip("\n")
    paragraphs = txt.split("\n\n")
    return [
        "\n".join(_clean(line) for line in paragraph.split("\n"))
        for paragraph in paragraphs
    ]


# Theme configuration - theme.base support functions


def _check_color_value(value: Any, option_name: str) -> None:
    """Validate theme color configuration option values.

    Validates that the value is a string (or list of strings, in the case of
    chartCategoricalColors and chartSequentialColors) and is not empty.

    Handles both single color strings (like primaryColor, backgroundColor)
    and arrays of color strings (like chartCategoricalColors, chartSequentialColors).

    Parameters
    ----------
    value : Any
        The color value to validate. Can be a string or list of strings.
    option_name : str
        The name of the theme option being validated (e.g., "theme.primaryColor").

    Raises
    ------
    StreamlitInvalidThemeOptionError
        If the value is not a string/list of strings, is empty, or contains
        empty values in the case of arrays.

    Notes
    -----
    Logs warnings for potentially invalid colors, since more comprehensive
    validation happens on the frontend.
    """
    logger = _get_logger()

    # Handle array color options (chartCategoricalColors, chartSequentialColors)
    if isinstance(value, list):
        if not value:
            raise StreamlitInvalidThemeOptionError(
                f"Theme option '{option_name}' cannot be an empty array"
            )

        for i, color in enumerate(value):
            if not isinstance(color, str):
                raise StreamlitInvalidThemeOptionError(
                    f"Theme option '{option_name}[{i}]' must be a string, got {type(color).__name__}: {color}"
                )

            color_str = color.strip()
            if not color_str:
                raise StreamlitInvalidThemeOptionError(
                    f"Theme option '{option_name}[{i}]' cannot be empty"
                )

            # Lightweight color validation with warning
            if not is_css_color_like(color_str):
                logger.warning(
                    "Theme option '%s[%s]' may be an invalid color: %s. "
                    "Expected formats: hex, rgb, and rgba colors",
                    option_name,
                    i,
                    color_str,
                )

        return  # All colors in array have been checked

    # Handle single color options (primaryColor, backgroundColor, etc.)
    if not isinstance(value, str):
        raise StreamlitInvalidThemeOptionError(
            f"Theme option '{option_name}' must be a string or array of strings, got {type(value).__name__}: {value}"
        )

    value_str: str = value.strip()

    if not value_str:
        raise StreamlitInvalidThemeOptionError(
            f"Theme option '{option_name}' cannot be empty"
        )

    # Lightweight color validation with warning
    if not is_css_color_like(value_str):
        logger.warning(
            "Theme option '%s' may be an invalid color: %s. "
            "Expected formats: hex, rgb, and rgba colors",
            option_name,
            value_str,
        )


def _iterate_theme_config_options(
    config_options: dict[str, ConfigOption],
) -> Iterator[tuple[str, Any]]:
    """
    Iterate through theme config options, yielding (option_path, value) pairs.
    Returns: theme.primaryColor, #ff0000, ...

    Leveraged by _extract_current_theme_config() to retrieve main config.toml theme options.
    """
    for opt_name, opt_val in config_options.items():
        if opt_name.startswith("theme.") and opt_val.value is not None:
            yield opt_name, opt_val.value


def _extract_current_theme_config(
    config_options: dict[str, ConfigOption],
) -> dict[str, Any]:
    """
    Extract current theme configuration from config options.
    Returns a dictionary with the current theme options in nested format.
    """
    current_theme_options = {}

    for opt_name, opt_value in _iterate_theme_config_options(config_options):
        parts = opt_name.split(".")
        if len(parts) == 2:  # theme.option
            _, option = parts
            if option != "base":  # Don't include the base option itself
                current_theme_options[option] = opt_value
        elif len(parts) == 3:  # theme.sidebar.option or theme.light.option
            _, section, option = parts
            if section not in current_theme_options:
                current_theme_options[section] = {}
            current_theme_options[section][option] = opt_value
        elif len(parts) == 4:  # theme.light.sidebar.option or theme.dark.sidebar.option
            _, section, subsection, option = parts
            if section not in current_theme_options:
                current_theme_options[section] = {}
            if subsection not in current_theme_options[section]:
                current_theme_options[section][subsection] = {}
            current_theme_options[section][subsection][option] = opt_value

    return current_theme_options


def _get_valid_theme_options(
    config_options_template: dict[str, ConfigOption],
) -> tuple[set[str], set[str]]:
    """Get valid theme configuration options for main theme and theme sections.

    Extracts valid theme options from the config options template to ensure they
    stay in sync with the actual theme options defined via _create_theme_options() calls.

    Parameters
    ----------
    config_options_template : dict[str, ConfigOption]
        Template of all available configuration options.

    Returns
    -------
    tuple[set[str], set[str]]
        A tuple (main_theme_options, section_theme_options) where:
        - main_theme_options: Valid theme options for the main theme (without "theme." prefix)
        - section_theme_options: Valid theme options for sections/subsections
          (sidebar, light, dark, light.sidebar, dark.sidebar)

    Notes
    -----
    All non-main theme sections have the same valid options, so we only need to
    extract them once.
    """
    # Extract options dynamically from the config template
    main_theme_options = set()
    section_theme_options = set()

    # Extract theme options from the config template
    for option_key in config_options_template:
        if option_key.startswith("theme."):
            parts = option_key.split(".")
            # Direct theme options like "theme.primaryColor"
            if parts[0] == "theme" and len(parts) == 2:
                _, option_name = parts
                main_theme_options.add(option_name)
            # Subsection options like "theme.sidebar.primaryColor"
            elif parts[0] == "theme" and parts[1] == "sidebar" and len(parts) == 3:
                # All subsections (sidebar, light, dark, light.sidebar, dark.sidebar)
                # get the same options as theme.sidebar (which excludes main-only options)
                _, _, option_name = parts
                section_theme_options.add(option_name)

    return main_theme_options, section_theme_options


def _invalid_theme_option_warning(
    option_name: str,
    file_path_or_url: str,
    valid_options: set[str],
    section_name: str = "theme",
) -> None:
    """Helper function to log a warning for an invalid theme option."""

    if section_name == "theme":
        full_option_name = f"{section_name}.{option_name}"
    else:
        # Handle sections like "sidebar" -> "theme.sidebar.{option_name}"
        # or subsections like "light.sidebar" -> "theme.light.sidebar.{option_name}"
        full_option_name = f"theme.{section_name}.{option_name}"

    valid_options_list = "\n".join(f"  • {opt}" for opt in sorted(valid_options))
    _get_logger().warning(
        "Theme file %s contains invalid theme option: '%s'.\n\n"
        "Valid '%s' options are:\n%s",
        file_path_or_url,
        full_option_name,
        section_name,
        valid_options_list,
    )


def _validate_theme_section_recursive(
    section_configs: dict[str, Any],
    section_path: str,
    file_path_or_url: str,
    section_options: set[str],
    filtered_parent: dict[str, Any],
    allow_sidebar_subsection: bool = False,
) -> None:
    """Recursively validate a theme section and its subsection/options.

    Parameters
    ----------
    section_configs : dict[str, Any]
        The section configs to validate.
    section_path : str
        Path like 'sidebar', 'light', 'light.sidebar'.
    file_path_or_url : str
        Theme file path for error messages.
    section_options : set[str]
        Valid options for this section.
    filtered_parent : dict[str, Any]
        Parent section to populate/filter out invalid options.
    allow_sidebar_subsection : bool, optional
        Allow sidebar subsection (only "light" and "dark" sections), by default False.

    Raises
    ------
    StreamlitInvalidThemeSectionError
        If an invalid subsection is found.
    """
    for option_name, option_value in section_configs.items():
        if isinstance(option_value, dict):
            # This is a subsection
            if not allow_sidebar_subsection or option_name != "sidebar":
                raise StreamlitInvalidThemeSectionError(
                    f"theme.{section_path}.{option_name}",
                    file_path_or_url,
                )

            # Create and validate the subsection's options
            if option_name not in filtered_parent:
                filtered_parent[option_name] = {}

            _validate_theme_section_recursive(
                option_value,
                f"{section_path}.{option_name}",
                file_path_or_url,
                section_options,
                filtered_parent[option_name],
                False,  # sidebar subsection can't have further subsections
            )
        elif option_name not in section_options:
            # This is an invalid section option
            _invalid_theme_option_warning(
                option_name,
                file_path_or_url,
                section_options,
                section_path,
            )
            # Remove the invalid option from the filtered theme
            filtered_parent.pop(option_name, None)
        else:
            # Valid option - add to filtered theme and check color values
            filtered_parent[option_name] = option_value
            full_option_name = f"theme.{section_path}.{option_name}"
            if "color" in full_option_name.lower():
                _check_color_value(option_value, full_option_name)


def _validate_theme_file_content(
    theme_content: dict[str, Any],
    file_path_or_url: str,
    config_options_template: dict[str, ConfigOption],
) -> dict[str, Any]:
    """
    Validate that a theme file contains only valid theme sections and config options.

    If invalid sections are found in the theme file, a StreamlitInvalidThemeSectionError is raised.

    If invalid config options are found in the theme file, a warning is logged with the valid
    options for the given section.

    Returns
    -------
        A filtered copy of the theme content with invalid options removed.
    """
    # Get valid options for each type of section
    valid_main_options, valid_section_options = _get_valid_theme_options(
        config_options_template
    )
    # Valid theme sections
    valid_sections = {"sidebar", "light", "dark"}

    theme_section = theme_content.get("theme", {})

    # Create a filtered copy of the theme content
    filtered_theme = copy.deepcopy(theme_content)
    filtered_theme_section = filtered_theme.get("theme", {})

    # Validate theme options
    for option_name, option_value in theme_section.items():
        # This is a section like theme.sidebar, theme.light, theme.dark
        if isinstance(option_value, dict):
            # Invalid section: raise error
            if option_name not in valid_sections:
                raise StreamlitInvalidThemeSectionError(
                    option_name,
                    file_path_or_url,
                )

            # Create the section in our filtered theme and validate it
            if option_name not in filtered_theme_section:
                filtered_theme_section[option_name] = {}

            # Subsection can only be sidebar from within light and dark sections
            allow_sidebar_subsection = option_name in {"light", "dark"}

            _validate_theme_section_recursive(
                option_value,
                option_name,
                file_path_or_url,
                valid_section_options,
                filtered_theme_section[option_name],
                allow_sidebar_subsection,
            )

        elif option_name not in valid_main_options:
            # Invalid main theme option
            _invalid_theme_option_warning(
                option_name,
                file_path_or_url,
                valid_main_options,
            )
            # Remove the invalid option from the filtered theme
            filtered_theme_section.pop(option_name, None)

        else:
            # Valid main theme option - if color config, check color value
            full_option_name = f"theme.{option_name}"
            if "color" in full_option_name.lower():
                _check_color_value(option_value, full_option_name)

    return filtered_theme


def _load_theme_file(
    file_path_or_url: str, config_options_template: dict[str, ConfigOption]
) -> dict[str, Any]:
    """
    Load and parse a theme TOML file from a local path or URL.

    Handles raising errors when a file cannot be found, read, parsed,
    or contains invalid theme options.

    Otherwise returns the parsed TOML content as a dictionary.
    """

    def _raise_missing_toml() -> None:
        raise StreamlitInvalidThemeError(
            "The 'toml' package is required to load theme files. "
            "Please install it with 'pip install toml'."
        )

    def _raise_file_not_found() -> None:
        raise FileNotFoundError(f"Theme file not found: {file_path_or_url}")

    def _raise_missing_theme_section() -> None:
        raise StreamlitInvalidThemeSectionError(
            f"Theme file {file_path_or_url} must contain a [theme] section"
        )

    def _raise_file_too_large() -> None:
        content_size = len(content.encode("utf-8"))
        raise StreamlitInvalidThemeError(
            f"Theme file {file_path_or_url} is too large ({content_size:,} bytes). "
            f"Maximum allowed size is {_MAX_THEME_FILE_SIZE_BYTES:,} bytes (1MB). "
            f"Theme files should contain only configuration options, not large data."
        )

    try:
        import toml
    except ImportError:
        _raise_missing_toml()

    # Check if it's a URL using the url_util helper (only allow http/https schemes by default)
    is_valid_url = url_util.is_url(file_path_or_url)

    try:
        if is_valid_url:
            # Load from URL - noqa: S310 suppressed since url_util.is_url() restricts to only
            # http/https schemes by default, preventing file:// or other dangerous schemes
            # 30-second timeout prevents hanging in poor network conditions (same as cli.py)
            with urllib.request.urlopen(file_path_or_url, timeout=30) as response:  # noqa: S310
                content = response.read().decode("utf-8")
        else:
            # Load from local file path
            # Resolve relative paths from the current working directory
            if not os.path.isabs(file_path_or_url):
                file_path_or_url = os.path.join(os.getcwd(), file_path_or_url)

            if not os.path.exists(file_path_or_url):
                _raise_file_not_found()

            with open(file_path_or_url, encoding="utf-8") as f:
                content = f.read()

        # Check file size limit - theme files should be small configuration files
        content_size = len(content.encode("utf-8"))
        if content_size > _MAX_THEME_FILE_SIZE_BYTES:
            _raise_file_too_large()

        # Parse the TOML content
        parsed_theme = toml.loads(content)  # ty: ignore[possibly-unresolved-reference]

        # Validate that the theme file has a theme section
        if "theme" not in parsed_theme:
            _raise_missing_theme_section()

        # Validate that the theme file contains only valid theme options, filtering out invalid ones
        filtered_theme = _validate_theme_file_content(
            parsed_theme, file_path_or_url, config_options_template
        )

        return filtered_theme

    except (
        StreamlitInvalidThemeError,
        StreamlitInvalidThemeOptionError,
        StreamlitInvalidThemeSectionError,
        FileNotFoundError,
    ):
        # Re-raise these specific exceptions
        raise
    except urllib.error.URLError as e:
        raise StreamlitInvalidThemeError(
            f"Could not load theme file from URL {file_path_or_url}: {e}"
        ) from e
    except Exception as e:
        raise StreamlitInvalidThemeError(
            f"Error loading theme file {file_path_or_url}: {e}"
        ) from e


def _deep_merge_theme_dicts(
    base_dict: dict[str, Any], override_dict: dict[str, Any]
) -> dict[str, Any]:
    """
    Recursively merge two dictionaries, with override_dict values taking precedence.
    Handles arbitrary levels of nesting for theme configurations.
    """
    merged = copy.deepcopy(base_dict)

    for key, value in override_dict.items():
        if key in merged and isinstance(merged[key], dict) and isinstance(value, dict):
            # Both base and override have dict values for this key, merge recursively
            merged[key] = _deep_merge_theme_dicts(merged[key], value)
        else:
            # Override value takes precedence (either new key or non-dict value)
            merged[key] = copy.deepcopy(value)

    return merged


def _apply_theme_inheritance(
    base_theme: dict[str, Any], override_theme: dict[str, Any]
) -> dict[str, Any]:
    """
    Apply theme inheritance where theme config values from config.toml
    take precedence over the theme configs defined in theme.base toml file.

    Returns a dictionary with the merged theme configuration.
    """
    return _deep_merge_theme_dicts(base_theme, override_theme)


def _set_theme_options_recursive(
    options_dict: dict[str, Any], prefix: str, set_option_func: Any, source: str
) -> None:
    """
    Recursively set theme options from nested dictionary in process_theme_inheritance().
    This utility function traverses nested theme configuration sections/subsection
    and sets each option using the provided set_option_func.
    """
    for option_name, option_value in options_dict.items():
        if option_name == "base" and prefix == "theme":
            # Base is handled separately in theme inheritance
            continue

        current_key = f"{prefix}.{option_name}" if prefix else option_name

        if isinstance(option_value, dict):
            # Recursively handle nested sections
            _set_theme_options_recursive(
                option_value, current_key, set_option_func, source
            )
        else:
            # Set the actual config option
            set_option_func(current_key, option_value, source)


# Theme configuration - handles theme.base


def process_theme_inheritance(
    config_options: dict[str, ConfigOption] | None,
    config_options_template: dict[str, ConfigOption],
    set_option_func: Any,
    config_file_paths: Sequence[str] | None = None,
) -> None:
    """
    Process theme inheritance if theme.base points to a theme file.

    This function checks if theme.base is set to a file path or URL,
    loads the theme file, and applies inheritance logic where the
    current config.toml values override the theme.base file values.

    Sets the merged theme options to the config.

    Parameters
    ----------
    config_options : dict[str, ConfigOption] | None
        Current configuration options.
    config_options_template : dict[str, ConfigOption]
        Template of all valid config options.
    set_option_func : Any
        Function to set config option values.
    config_file_paths : Sequence[str] | None, optional
        Ordered list of config.toml file paths (global, project, script).
        Used to label where_defined with scope for clearer provenance.
    """
    # Get the current theme.base value
    if config_options is None:
        return

    base_option = config_options.get("theme.base")
    if not base_option or base_option.value is None:
        return

    base_value = base_option.value

    # Check if it's a file path or URL (not just "light" or "dark")
    if base_value in {"light", "dark"}:
        return

    def _raise_invalid_nested_base() -> None:
        raise StreamlitInvalidThemeError(
            f"Theme file {base_value} cannot reference another theme file in its base property. "
            f"Only 'light' and 'dark' are allowed in referenced theme files."
        )

    try:
        # Load the theme file config options
        theme_file_content = _load_theme_file(base_value, config_options_template)

        # Validate that theme.base of the referenced theme file doesn't reference another file
        theme_base = theme_file_content.get("theme", {}).get("base")
        if theme_base and theme_base not in {"light", "dark"}:
            _raise_invalid_nested_base()

        # Get current theme options from main config.toml
        current_theme_options = (
            _extract_current_theme_config(config_options) if config_options else {}
        )

        # Apply inheritance: referenced theme file as base, override with theme options specified in config.toml
        merged_theme = _apply_theme_inheritance(
            theme_file_content, {"theme": current_theme_options}
        )

        # Preserve theme options set by env vars and command line flags (higher precedence)
        high_precedence_theme_options: dict[str, _ThemeOverride] = {}
        config_theme_overrides: dict[str, _ThemeOverride] = {}
        if config_options is not None:
            for opt_name, opt_config in config_options.items():
                if (
                    opt_name.startswith("theme.")
                    and opt_name != "theme.base"
                    and opt_config.where_defined
                    in {
                        "environment variable",
                        "command-line argument or environment variable",
                    }
                ):
                    high_precedence_theme_options[opt_name] = {
                        "value": opt_config.value,
                        "where_defined": opt_config.where_defined,
                    }
                config_label = _format_config_file_label(
                    opt_config.where_defined, config_file_paths
                )
                if (
                    opt_name.startswith("theme.")
                    and opt_name != "theme.base"
                    and config_label is not None
                ):
                    config_theme_overrides[opt_name] = {
                        "value": opt_config.value,
                        "where_defined": config_label,
                    }

            # Clear existing theme options (except base) to prepare for inheritance
            theme_options_to_remove = [
                opt_name
                for opt_name in config_options
                if opt_name.startswith("theme.") and opt_name != "theme.base"
            ]
            for opt_name in theme_options_to_remove:
                set_option_func(opt_name, None, "reset for theme inheritance")

        # Handle theme.base - always set it to a valid value ("light" or "dark", not a path/URL)
        theme_file_base = theme_file_content.get("theme", {}).get("base")
        if theme_file_base:
            set_option_func(
                "theme.base", theme_file_base, f"base theme file: {base_value}"
            )
        else:
            # Theme file doesn't specify a base, default to "light"
            set_option_func("theme.base", "light", "default light theme")

        # Set the merged theme options using recursive helper
        theme_section = merged_theme.get("theme", {})
        _set_theme_options_recursive(
            theme_section, "theme", set_option_func, f"base theme file: {base_value}"
        )

        # Restore theme options set in config.toml (override base theme file)
        for opt_name, opt_data in config_theme_overrides.items():
            set_option_func(opt_name, opt_data["value"], opt_data["where_defined"])

        # Finally, restore theme options set by env vars and command line flags (highest precedence)
        for opt_name, opt_data in high_precedence_theme_options.items():
            set_option_func(opt_name, opt_data["value"], opt_data["where_defined"])

    except (
        StreamlitInvalidThemeError,
        StreamlitInvalidThemeOptionError,
        StreamlitInvalidThemeSectionError,
        FileNotFoundError,
    ):
        # Re-raise expected user errors as-is to preserve specific error messages
        raise
    except Exception as e:
        _get_logger().exception("Error processing theme inheritance")
        # Only wrap unexpected errors (not our specific validation errors)
        raise StreamlitInvalidThemeError(
            f"Failed to process theme inheritance from {base_value}: {e}"
        ) from e



# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metrics dashboard template.

Demonstrates:
- Time series visualization with Altair (line, area, bar, point charts)
- Metric cards with chart/table toggle and popover filters
- Time range filtering (1M, 6M, 1Y, QTD, YTD, All)
- Line options (Daily, 7-day MA)

This template uses synthetic data. Replace the ``generate_*_data()`` functions
with your own data sources (e.g., Snowflake queries, APIs, etc.).
"""

import hashlib
from datetime import date, timedelta

import altair as alt
import numpy as np
import pandas as pd

import streamlit as st

st.set_page_config(
    page_title="Metrics Dashboard",
    page_icon=":material/monitoring:",
    layout="wide",
)


# =============================================================================
# Constants
# =============================================================================

TIME_RANGES = ["1M", "6M", "1Y", "QTD", "YTD", "All"]
CHART_HEIGHT = 300


# =============================================================================
# Synthetic Data Generation (Replace with your data source)
# =============================================================================


def generate_metric_data(
    metric_name: str,
    start_date: date,
    end_date: date,
    base_value: float = 1000,
    growth_rate: float = 0.001,
    noise_factor: float = 0.1,
) -> pd.DataFrame:
    """Generate synthetic time series data for a metric.

    Replace this function with your actual data source, e.g.:
    - Snowflake query
    - API call
    - Database query
    """
    seed = int(hashlib.sha256(metric_name.encode()).hexdigest(), 16) % 2**32
    rng = np.random.default_rng(seed)

    dates = pd.date_range(start=start_date, end=end_date, freq="D")
    n_days = len(dates)

    # Base trend with growth
    trend = base_value * (1 + growth_rate) ** np.arange(n_days)

    # Add weekly seasonality (lower on weekends)
    day_of_week = dates.dayofweek
    seasonality = np.where(day_of_week >= 5, 0.7, 1.0)
    trend *= seasonality

    # Add noise
    noise = rng.normal(1, noise_factor, n_days)
    values = trend * noise

    # Calculate rolling averages
    df = pd.DataFrame(
        {
            "ds": dates,
            "daily_value": values,
        }
    )
    df["value_7d_ma"] = df["daily_value"].rolling(7, min_periods=1).mean()

    return df


@st.cache_data(ttl=3600)
def load_all_metrics() -> dict[str, pd.DataFrame]:
    """Load all metrics data. Replace with your data loading logic."""
    end_date = date.today() - timedelta(days=1)
    start_date = end_date - timedelta(days=730)  # 2 years of data

    return {
        "users": generate_metric_data(
            "users", start_date, end_date, base_value=5000, growth_rate=0.002
        ),
        "sessions": generate_metric_data(
            "sessions", start_date, end_date, base_value=15000, growth_rate=0.003
        ),
        "revenue": generate_metric_data(
            "revenue", start_date, end_date, base_value=50000, growth_rate=0.001
        ),
        "conversions": generate_metric_data(
            "conversions", start_date, end_date, base_value=500, growth_rate=0.0015
        ),
    }


# =============================================================================
# Chart Utilities
# =============================================================================


def filter_by_time_range(df: pd.DataFrame, x_col: str, time_range: str) -> pd.DataFrame:
    """Filter dataframe by time range."""
    if time_range == "All" or df.empty:
        return df

    df = df.copy()
    df[x_col] = pd.to_datetime(df[x_col])
    max_date = df[x_col].max()

    if time_range == "1M":
        min_date = max_date - timedelta(days=30)
    elif time_range == "6M":
        min_date = max_date - timedelta(days=180)
    elif time_range == "1Y":
        min_date = max_date - timedelta(days=365)
    elif time_range == "QTD":
        quarter_month = ((max_date.month - 1) // 3) * 3 + 1
        min_date = pd.Timestamp(date(max_date.year, quarter_month, 1))
    elif time_range == "YTD":
        min_date = pd.Timestamp(date(max_date.year, 1, 1))
    else:
        return df

    filtered: pd.DataFrame = df[df[x_col] >= min_date]
    return filtered


def _melt_chart_data(
    df: pd.DataFrame, x_col: str, y_cols: list[str], labels: list[str]
) -> pd.DataFrame:
    """Melt wide data into long form and re-label the series column."""
    melted = df.melt(
        id_vars=[x_col],
        value_vars=y_cols,
        var_name="series",
        value_name="value",
    )
    label_map = dict(zip(y_cols, labels, strict=True))
    melted["series"] = melted["series"].map(label_map)
    return melted


def _chart_tooltip(x_col: str, x_title: str = "Date") -> list[alt.Tooltip]:
    """Standard tooltip for melted chart data."""
    return [
        alt.Tooltip(f"{x_col}:T", title=x_title, format="%Y-%m-%d"),
        alt.Tooltip("series:N", title="Series"),
        alt.Tooltip("value:Q", title="Value", format=",.0f"),
    ]


def render_line_chart(
    df: pd.DataFrame,
    x_col: str,
    y_cols: list[str],
    labels: list[str],
    height: int = CHART_HEIGHT,
) -> alt.Chart:
    """Render a multi-line chart."""
    melted = _melt_chart_data(df, x_col, y_cols, labels)

    chart: alt.Chart = (
        alt.Chart(melted)
        .mark_line()
        .encode(
            x=alt.X(f"{x_col}:T", title=None),
            y=alt.Y("value:Q", title=None, scale=alt.Scale(zero=False)),
            color=alt.Color("series:N", title=None, legend=alt.Legend(orient="bottom")),
            strokeDash=alt.condition(
                alt.datum.series == "7-day MA",
                alt.value([5, 5]),
                alt.value([0]),
            ),
            tooltip=_chart_tooltip(x_col),
        )
        .properties(height=height)
    )
    return chart


def render_area_chart(
    df: pd.DataFrame,
    x_col: str,
    y_cols: list[str],
    labels: list[str],
    height: int = CHART_HEIGHT,
) -> alt.Chart:
    """Render a stacked area chart."""
    melted = _melt_chart_data(df, x_col, y_cols, labels)

    chart: alt.Chart = (
        alt.Chart(melted)
        .mark_area(opacity=0.6, line=True)
        .encode(
            x=alt.X(f"{x_col}:T", title=None),
            y=alt.Y("value:Q", title=None, scale=alt.Scale(zero=False)),
            color=alt.Color("series:N", title=None, legend=alt.Legend(orient="bottom")),
            tooltip=_chart_tooltip(x_col),
        )
        .properties(height=height)
    )
    return chart


def render_bar_chart(
    df: pd.DataFrame,
    x_col: str,
    y_cols: list[str],
    labels: list[str],
    height: int = CHART_HEIGHT,
) -> alt.Chart:
    """Render a bar chart (weekly aggregation for readability)."""
    df = df.copy()
    df[x_col] = pd.to_datetime(df[x_col])
    df["week"] = df[x_col].dt.to_period("W").dt.start_time

    agg_df = df.groupby("week")[y_cols].mean().reset_index()
    melted = _melt_chart_data(agg_df, "week", y_cols, labels)

    chart: alt.Chart = (
        alt.Chart(melted)
        .mark_bar(opacity=0.8)
        .encode(
            x=alt.X("week:T", title=None),
            y=alt.Y("value:Q", title=None, scale=alt.Scale(zero=False)),
            color=alt.Color("series:N", title=None, legend=alt.Legend(orient="bottom")),
            xOffset="series:N",
            tooltip=_chart_tooltip("week", x_title="Week"),
        )
        .properties(height=height)
    )
    return chart


def render_point_chart(
    df: pd.DataFrame,
    x_col: str,
    y_cols: list[str],
    labels: list[str],
    height: int = CHART_HEIGHT,
) -> alt.Chart:
    """Render a scatter/point chart with a dashed trend line for the 7-day MA."""
    melted = _melt_chart_data(df, x_col, y_cols, labels)

    points = (
        alt.Chart(melted)
        .mark_point(opacity=0.5, size=20)
        .encode(
            x=alt.X(f"{x_col}:T", title=None),
            y=alt.Y("value:Q", title=None, scale=alt.Scale(zero=False)),
            color=alt.Color("series:N", title=None, legend=alt.Legend(orient="bottom")),
            tooltip=_chart_tooltip(x_col),
        )
    )

    trend = (
        alt.Chart(melted[melted["series"] == "7-day MA"])
        .mark_line(strokeDash=[5, 5], strokeWidth=2)
        .encode(
            x=alt.X(f"{x_col}:T"),
            y=alt.Y("value:Q"),
            color=alt.Color("series:N"),
        )
    )

    chart: alt.Chart = (points + trend).properties(height=height)
    return chart


# =============================================================================
# Metric Card Component
# =============================================================================


def metric_card(
    title: str,
    df: pd.DataFrame,
    key_prefix: str,
    chart_type: str = "line",
) -> None:
    """Display a metric card with chart/table toggle and popover filters.

    Parameters
    ----------
    title : str
        Card title.
    df : pd.DataFrame
        DataFrame with ``ds``, ``daily_value``, ``value_7d_ma`` columns.
    key_prefix : str
        Unique prefix for widget keys.
    chart_type : str
        One of ``"line"``, ``"area"``, ``"bar"``, ``"point"``.
    """
    chart_renderers = {
        "line": render_line_chart,
        "area": render_area_chart,
        "bar": render_bar_chart,
        "point": render_point_chart,
    }
    render_chart = chart_renderers.get(chart_type, render_line_chart)

    with st.container(border=True):
        # Header row with title, view toggle, and filters
        with st.container(
            horizontal=True,
            horizontal_alignment="distribute",
            vertical_alignment="center",
        ):
            st.markdown(f"**{title}**")

            view_mode = st.segmented_control(
                "View",
                options=[":material/show_chart:", ":material/table:"],
                default=":material/show_chart:",
                key=f"{key_prefix}_view",
                label_visibility="collapsed",
            )

            with st.popover("Filters", type="tertiary"):
                line_options = st.pills(
                    "Lines",
                    options=["Daily", "7-day MA"],
                    default=["Daily", "7-day MA"],
                    selection_mode="multi",
                    key=f"{key_prefix}_lines",
                )
                time_range = st.segmented_control(
                    "Time range",
                    options=TIME_RANGES,
                    default="All",
                    key=f"{key_prefix}_time",
                )

        # Apply filters
        line_options = line_options or ["7-day MA"]
        filtered_df = filter_by_time_range(df, "ds", time_range or "All")

        # Determine which columns to show
        y_cols = []
        labels = []
        if "Daily" in line_options:
            y_cols.append("daily_value")
            labels.append("Daily")
        if "7-day MA" in line_options:
            y_cols.append("value_7d_ma")
            labels.append("7-day MA")

        # Render view
        if "table" in (view_mode or ""):
            st.dataframe(
                filtered_df,
                height=CHART_HEIGHT,
                hide_index=True,
            )
        elif y_cols:
            st.altair_chart(
                render_chart(filtered_df, "ds", y_cols, labels),
            )
        else:
            st.info("Select at least one line option.")


# =============================================================================
# Page Header Component
# =============================================================================


def render_page_header(title: str) -> None:
    """Render page header with title and reset button."""
    with st.container(
        horizontal=True, horizontal_alignment="distribute", vertical_alignment="center"
    ):
        st.markdown(title)
        if st.button(":material/restart_alt: Reset", type="tertiary"):
            st.session_state.clear()
            st.rerun()


# =============================================================================
# Page Layout
# =============================================================================

# Load data (cached)
metrics_data = load_all_metrics()

# Page header
render_page_header("# :material/monitoring: Metrics Dashboard")

# Row 1: Users and Sessions
row1 = st.columns(2)

with row1[0]:
    metric_card("Active Users", metrics_data["users"], "users", chart_type="line")

with row1[1]:
    metric_card("Sessions", metrics_data["sessions"], "sessions", chart_type="area")

# Row 2: Revenue and Conversions
row2 = st.columns(2)

with row2[0]:
    metric_card("Revenue", metrics_data["revenue"], "revenue", chart_type="bar")

with row2[1]:
    metric_card(
        "Conversions", metrics_data["conversions"], "conversions", chart_type="point"
    )



# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared typing utilities for the `st.components.v2` API.

This module exposes common, user-facing argument types and callable
signatures used by the bidirectional component API. Import these types to
annotate code that constructs kwargs dictionaries for components, or when
authoring wrappers/utilities around `st.components.v2.component`.

The goal is to keep the public argument surface documented in one place and
reusable across both the user-facing factory in `components/v2/__init__.py`
and the internal implementation in `components/v2/bidi_component/main.py`.

Component definitions may provide any combination of HTML, CSS, and
JavaScript. If none are supplied, the component renders as an empty
placeholder without raising an error.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:
    from streamlit.components.v2.bidi_component.state import ComponentResult
    from streamlit.elements.lib.layout_utils import Height, Width
    from streamlit.runtime.state.common import WidgetCallback


# Individual argument type aliases to make reuse ergonomic across modules.
BidiComponentKey = str | None
BidiComponentData = Any | None
BidiComponentDefaults = dict[str, Any] | None
ComponentIsolateStyles = bool


class ComponentRenderer(Protocol):
    '''Signature of the mounting command returned by ``st.components.v2.component``.

    This callable mounts a bidirectional component in a Streamlit app and
    returns a ``ComponentResult`` object that exposes the component's
    state and trigger values.

    For published components, this callable is often wrapped in a user-friendly
    command with typed parameters and declared defaults.

    Parameters
    ----------
    key : str or None
        An optional string to use as the unique key for the
        component instance. If this is omitted, an internal key is generated
        for the component instance based on its mounting parameters. No two
        Streamlit elements may have the same key.

        When a key is defined, the component's state is available in Session
        State via the key.

        .. note::
            If you want to access this key in your component's frontend, you
            must pass it explicitly within the ``data`` parameter. The ``key``
            parameter in ``ComponentRenderer`` is not the same as the
            ``key`` property in ``FrontendRendererArgs`` in the component's frontend
            code.

            The frontend key is automatically generated to be unique among all
            instances of all components and to avoid collisions with classes
            and IDs in the app's DOM.

    data : Any or None
        Data to pass to the component. This can be one of the following:

        - A JSON-serializable object, like ``Dict[str, str | int]`` or
          ``List[str]``.
        - An Arrow-serializable object, like ``pandas.DataFrame``.
        - Raw bytes.
        - A dictionary of JSON-serializable and Arrow-serializable objects.
          The dictionary's keys must be Python primitives.

        Because this data is sent to the frontend, it must be serializable by
        one of the supported serialization methods (JSON, Arrow, or raw bytes).
        You can't pass arbitrary Python objects. Arrow-serialization is only
        supported at the top level of the ``data`` parameter or one level deep
        in a dictionary. Raw bytes are only supported at the top level.

    default : dict[str, Any] or None
        Default state values for the component. Each key in the dictionary must
        correspond to a valid state attribute with an ``on_<key>_change``
        callback. This callback can be empty, but must be included as a
        parameter when the component is mounted.

        Trigger values do not support manual defaults. All trigger and state
        values defined by an associated callback are initialized to ``None`` by
        default.

    width : "stretch", "content", or int
        Width of the component. This can be one of the following:

        - ``"stretch"`` (default): The component is wrapped in a ``<div>`` with
          CSS style ``width: 100%;``.
        - ``"content"``: The component is wrapped in a ``<div>`` with CSS
          style ``width: fit-content;``.
        - An integer specifying the width in pixels: The component is wrapped
          in a ``<div>`` with the specified pixel width.

        You are responsible for ensuring the component's inner HTML content is
        responsive to the ``<div>`` wrapper.

    height : "content", "stretch", or int
        Height of the component. This can be one of the following:

        - ``"content"`` (default): The component is wrapped in a ``<div>`` with
          CSS style ``height: auto;``.
        - ``"stretch"``: The component is wrapped in a ``<div>`` with CSS
          style ``height: 100%;``.
        - An integer specifying the height in pixels: The component is wrapped
          in a ``<div>`` with the specified pixel height. If the component
          content is larger than the specified height, scrolling is enabled.

        .. note::
            Use scrolling containers sparingly. If you use scrolling
            containers, avoid heights that exceed 500 pixels. Otherwise,
            the scroll surface of the container might cover the majority of
            the screen on mobile devices, which makes it hard to scroll the
            rest of the app.

            If you want to disable scrolling for a fixed-height component,
            include an inner ``<div>`` wrapper in your component's HTML to
            control the overflow behavior.

        You are responsible for ensuring the component's inner HTML content is
        responsive to the ``<div>`` wrapper.

    **callbacks : Callable or None
        Callbacks with the naming pattern ``on_<key>_change`` for each state and
        trigger key. For example, if your component has a state key of
        ``"value"`` and a trigger key of ``"click"``, its callbacks can include
        ``on_value_change`` and ``on_click_change``.

        Only names that follow this pattern are recognized. Custom components
        don't currently support callbacks with arguments.

        Callbacks are required for any state values defined in the ``default``
        parameter. Otherwise, a callback is optional. To ensure your
        component's result always returns the expected attributes, you can pass
        empty callbacks like ``lambda: None``.

    Returns
    -------
    ComponentResult
        Component state object that exposes state and trigger values.

    Examples
    --------
    **Example 1: Create a bidirectional text input component**

    If you assign a key to a mounted instance of a component, you can feed its
    state back into the component through the ``data`` parameter. This allows
    you to both read and write state values from Session State. The following
    example has a user-friendly wrapper around the mounting command to provide
    typed parameters and a clean end-user API. A couple buttons demonstrate
    programmatic updates to the component's state.

    .. code-block:: python

        import streamlit as st

        HTML = """
            <label style='padding-right: 1em;' for='txt'>Enter text</label>
            <input id='txt' type='text' />
        """

        JS = """
            export default function(component) {
                const { setStateValue, parentElement, data } = component;

                const label = parentElement.querySelector('label');
                label.innerText = data.label;

                const input = parentElement.querySelector('input');
                if (input.value !== data.value) {
                    input.value = data.value ?? '';
                };

                input.onkeydown = (e) => {
                    if (e.key === 'Enter') {
                        setStateValue('value', e.target.value);
                    }
                };

                input.onblur = (e) => {
                    setStateValue('value', e.target.value);
                };
            }
        """

        my_component = st.components.v2.component(
            "my_text_input",
            html=HTML,
            js=JS,
        )


        def my_component_wrapper(
            label, *, default="", key=None, on_change=lambda: None
        ):
            component_state = st.session_state.get(key, {})
            value = component_state.get("value", default)
            data = {"label": label, "value": value}
            result = my_component(
                data=data,
                default={"value": value},
                key=key,
                on_value_change=on_change,
            )
            return result


        st.title("My custom component")

        if st.button("Hello World"):
            st.session_state["my_text_input_instance"]["value"] = "Hello World"
        if st.button("Clear text"):
            st.session_state["my_text_input_instance"]["value"] = ""
        result = my_component_wrapper(
            "Enter something",
            default="I love Streamlit!",
            key="my_text_input_instance",
        )

        st.write("Result:", result)
        st.write("Session state:", st.session_state)

    .. output::
        https://doc-components-text-input.streamlit.app/
        height: 600px

    **Example 2: Add Tailwind CSS to a component**

    You can use the ``isolate_styles`` parameter in
    ``st.components.v2.component`` to disable shadow DOM isolation and apply
    global styles like Tailwind CSS to your component. The following example
    creates a simple button styled with Tailwind CSS. This example also
    demonstrates using different keys to mount multiple instances of the same
    component in one app.

    .. code-block:: python

        import streamlit as st

        with open("tailwind.js", "r") as f:
            TAILWIND_SCRIPT = f.read()

        HTML = """
            <button class="bg-blue-500 hover:bg-blue-700 text-white py-1 px-3 rounded">
                Click me!
            </button>
        """
        JS = (
            TAILWIND_SCRIPT
            + """
                export default function(component) {
                    const { setTriggerValue, parentElement } = component;
                    const button = parentElement.querySelector('button');
                    button.onclick = () => {
                        setTriggerValue('clicked', true);
                    };
                }
            """
        )
        my_component = st.components.v2.component(
            "my_tailwind_button",
            html=HTML,
            js=JS,
            isolate_styles=False,
        )
        result_1 = my_component(on_clicked_change=lambda: None, key="one")
        result_1

        result_2 = my_component(on_clicked_change=lambda: None, key="two")
        result_2

    .. output::
        https://doc-components-tailwind-button.streamlit.app/
        height: 350px

    '''

    def __call__(
        self,
        *,
        key: BidiComponentKey = None,
        data: BidiComponentData = None,
        default: BidiComponentDefaults = None,
        width: Width = "stretch",
        height: Height = "content",
        **on_callbacks: WidgetCallback | None,
    ) -> ComponentResult: ...


__all__ = [
    "BidiComponentData",
    "BidiComponentDefaults",
    "BidiComponentKey",
    "ComponentIsolateStyles",
    "ComponentRenderer",
]



# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import re
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta
from textwrap import dedent
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Literal,
    TypeAlias,
    cast,
    overload,
)

from streamlit.elements.lib.form_utils import current_form_id
from streamlit.elements.lib.layout_utils import (
    WidthWithoutContent,
    create_layout_config,
)
from streamlit.elements.lib.policies import (
    check_widget_policies,
    maybe_raise_label_warnings,
)
from streamlit.elements.lib.utils import (
    Key,
    LabelVisibility,
    compute_and_register_element_id,
    get_label_visibility_proto_value,
    to_key,
)
from streamlit.errors import StreamlitAPIException
from streamlit.proto.DateInput_pb2 import DateInput as DateInputProto
from streamlit.proto.DateTimeInput_pb2 import DateTimeInput as DateTimeInputProto
from streamlit.proto.TimeInput_pb2 import TimeInput as TimeInputProto
from streamlit.runtime.metrics_util import gather_metrics
from streamlit.runtime.scriptrunner import ScriptRunContext, get_script_run_ctx
from streamlit.runtime.state import (
    BindOption,
    WidgetArgs,
    WidgetCallback,
    WidgetKwargs,
    get_session_state,
    register_widget,
)
from streamlit.time_util import adjust_years

if TYPE_CHECKING:
    from streamlit.delta_generator import DeltaGenerator

# Type for things that point to a specific time (even if a default time, though not None).
TimeValue: TypeAlias = time | datetime | str | Literal["now"]
DateTimeScalarValue: TypeAlias = datetime | date | time | str | Literal["now"]
DateTimeValue: TypeAlias = DateTimeScalarValue | None

# Type for things that point to a specific date (even if a default date, including None).
NullableScalarDateValue: TypeAlias = date | datetime | str | Literal["today"] | None

# The accepted input value for st.date_input. Can be a date scalar or a date range.
DateValue: TypeAlias = NullableScalarDateValue | Sequence[NullableScalarDateValue]

# The return value of st.date_input.
DateWidgetRangeReturn: TypeAlias = tuple[()] | tuple[date] | tuple[date, date]
DateWidgetReturn: TypeAlias = date | DateWidgetRangeReturn | None


DEFAULT_STEP_MINUTES: Final = 15
ALLOWED_DATE_FORMATS: Final = re.compile(
    r"^(YYYY[/.\-]MM[/.\-]DD|DD[/.\-]MM[/.\-]YYYY|MM[/.\-]DD[/.\-]YYYY)$"
)
_DATETIME_LEGACY_FORMAT: Final = "%Y/%m/%d, %H:%M"
_DATETIME_ISO_FORMAT: Final = "%Y-%m-%dT%H:%M"
_DEFAULT_MIN_BOUND_TIME: Final = time(hour=0, minute=0)
_DEFAULT_MAX_BOUND_TIME: Final = time(hour=23, minute=59)


def _convert_timelike_to_time(value: TimeValue) -> time:
    if value == "now":
        # Set value default.
        return datetime.now().time().replace(second=0, microsecond=0)

    if isinstance(value, str):
        try:
            return time.fromisoformat(value)
        except ValueError:
            try:
                return (
                    datetime.fromisoformat(value)
                    .time()
                    .replace(second=0, microsecond=0)
                )
            except ValueError:
                # We throw an error below.
                pass

    if isinstance(value, datetime):
        return value.time().replace(second=0, microsecond=0)

    if isinstance(value, time):
        return value

    raise StreamlitAPIException(
        "The type of value should be one of datetime, time, ISO string or None"
    )


def _convert_datelike_to_date(
    value: NullableScalarDateValue,
) -> date:
    if isinstance(value, datetime):
        return value.date()

    if isinstance(value, date):
        return value

    if value == "today":
        return datetime.now().date()

    if isinstance(value, str):
        try:
            return date.fromisoformat(value)
        except ValueError:
            try:
                return datetime.fromisoformat(value).date()
            except ValueError:
                # We throw an error below.
                pass

    raise StreamlitAPIException(
        'Date value should either be an date/datetime or an ISO string or "today"'
    )


def _parse_date_value(value: DateValue) -> tuple[list[date] | None, bool]:
    if value is None:
        return None, False

    value_tuple: Sequence[NullableScalarDateValue]

    if isinstance(value, Sequence) and not isinstance(value, str):
        is_range = True
        value_tuple = value  # ty: ignore[invalid-assignment]
    else:
        is_range = False
        value_tuple = [cast("NullableScalarDateValue", value)]

    if len(value_tuple) not in {0, 1, 2}:
        raise StreamlitAPIException(
            "DateInput value should either be an date/datetime or a list/tuple of "
            "0 - 2 date/datetime values"
        )

    parsed_dates = [_convert_datelike_to_date(v) for v in value_tuple]

    return parsed_dates, is_range


def _parse_min_date(
    min_value: NullableScalarDateValue,
    parsed_dates: Sequence[date] | None,
) -> date:
    parsed_min_date: date
    if isinstance(min_value, (datetime, date, str)):
        parsed_min_date = _convert_datelike_to_date(min_value)
    elif min_value is None:
        if parsed_dates:
            parsed_min_date = adjust_years(parsed_dates[0], years=-10)
        else:
            parsed_min_date = adjust_years(date.today(), years=-10)
    else:
        raise StreamlitAPIException(
            "DateInput min should either be a date/datetime or None"
        )
    return parsed_min_date


def _parse_max_date(
    max_value: NullableScalarDateValue,
    parsed_dates: Sequence[date] | None,
) -> date:
    parsed_max_date: date
    if isinstance(max_value, (datetime, date, str)):
        parsed_max_date = _convert_datelike_to_date(max_value)
    elif max_value is None:
        if parsed_dates:
            parsed_max_date = adjust_years(parsed_dates[-1], years=10)
        else:
            parsed_max_date = adjust_years(date.today(), years=10)
    else:
        raise StreamlitAPIException(
            "DateInput max should either be a date/datetime or None"
        )
    return parsed_max_date


def _normalize_time(value: time) -> time:
    """Return a time without seconds, microseconds, or timezone info."""
    return value.replace(second=0, microsecond=0, tzinfo=None)


def _normalize_datetime_value(value: datetime) -> datetime:
    """Return a datetime without seconds, microseconds, or timezone info."""
    if value.tzinfo is not None:
        value = value.replace(tzinfo=None)
    return value.replace(second=0, microsecond=0)


def _combine_date_time(component_date: date, component_time: time) -> datetime:
    """Combine a date and time into a normalized datetime."""
    return datetime.combine(component_date, _normalize_time(component_time))


def _try_parse_datetime_with_format(value: str, fmt: str) -> datetime | None:
    """Try to parse a datetime string with a specific format."""
    try:
        return datetime.strptime(value, fmt)
    except ValueError:
        return None


def _convert_datetimelike_to_datetime(
    value: DateTimeScalarValue,
    *,
    fallback_date: date,
    fallback_time: time,
) -> datetime:
    """Convert supported datetime inputs into a normalized datetime."""
    fallback_time = _normalize_time(fallback_time)

    if value == "now":
        return _normalize_datetime_value(datetime.now())

    if isinstance(value, datetime):
        return _normalize_datetime_value(value)

    if isinstance(value, date) and not isinstance(value, datetime):
        return _combine_date_time(value, fallback_time)

    if isinstance(value, time):
        return _combine_date_time(fallback_date, value)

    if isinstance(value, str):
        stripped_value = value.strip()

        try:
            parsed_dt = datetime.fromisoformat(stripped_value)
            return _normalize_datetime_value(parsed_dt)
        except ValueError:
            pass

        for fmt in (
            "%Y/%m/%d %H:%M",
            "%Y/%m/%d %H:%M:%S",
            "%Y-%m-%d %H:%M",
            "%Y-%m-%d %H:%M:%S",
        ):
            maybe_parsed_dt = _try_parse_datetime_with_format(stripped_value, fmt)
            if maybe_parsed_dt is not None:
                return _normalize_datetime_value(maybe_parsed_dt)

        try:
            parsed_date = date.fromisoformat(stripped_value)
            return _combine_date_time(parsed_date, fallback_time)
        except ValueError:
            pass

        try:
            parsed_time = time.fromisoformat(stripped_value)
            return _combine_date_time(fallback_date, parsed_time)
        except ValueError:
            pass

    raise StreamlitAPIException(
        "The type of value should be one of datetime, date, time, ISO string, or 'now'."
    )


def _default_min_datetime(base_date: date) -> datetime:
    return _combine_date_time(
        adjust_years(base_date, years=-10), _DEFAULT_MIN_BOUND_TIME
    )


def _default_max_datetime(base_date: date) -> datetime:
    return _combine_date_time(
        adjust_years(base_date, years=10), _DEFAULT_MAX_BOUND_TIME
    )


def _datetime_to_proto_string(value: datetime) -> str:
    return _normalize_datetime_value(value).strftime(_DATETIME_ISO_FORMAT)


@dataclass(frozen=True)
class _DateTimeInputValues:
    value: datetime | None
    min: datetime
    max: datetime

    @classmethod
    def from_raw_values(
        cls,
        value: DateTimeValue,
        min_value: DateTimeValue,
        max_value: DateTimeValue,
    ) -> _DateTimeInputValues:
        parsed_value = (
            None
            if value is None
            else _convert_datetimelike_to_datetime(
                value,
                fallback_date=date.today(),
                fallback_time=_DEFAULT_MIN_BOUND_TIME,
            )
        )

        base_date_for_bounds = (
            parsed_value.date() if parsed_value is not None else date.today()
        )

        parsed_min = (
            _default_min_datetime(base_date_for_bounds)
            if min_value is None
            else _convert_datetimelike_to_datetime(
                min_value,
                fallback_date=base_date_for_bounds,
                fallback_time=_DEFAULT_MIN_BOUND_TIME,
            )
        )

        parsed_max = (
            _default_max_datetime(base_date_for_bounds)
            if max_value is None
            else _convert_datetimelike_to_datetime(
                max_value,
                fallback_date=base_date_for_bounds,
                fallback_time=_DEFAULT_MAX_BOUND_TIME,
            )
        )

        return cls(
            value=parsed_value,
            min=parsed_min,
            max=parsed_max,
        )

    def __post_init__(self) -> None:
        if self.min > self.max:
            raise StreamlitAPIException(
                f"The `min_value`, set to {self.min}, shouldn't be larger "
                f"than the `max_value`, set to {self.max}."
            )

        if self.value is not None and (self.value < self.min or self.value > self.max):
            raise StreamlitAPIException(
                f"The default `value` of {self.value} must lie between the `min_value` "
                f"of {self.min} and the `max_value` of {self.max}, inclusively."
            )


@dataclass(frozen=True)
class _DateInputValues:
    value: Sequence[date] | None
    is_range: bool
    max: date
    min: date

    @classmethod
    def from_raw_values(
        cls,
        value: DateValue,
        min_value: NullableScalarDateValue,
        max_value: NullableScalarDateValue,
    ) -> _DateInputValues:
        parsed_value, is_range = _parse_date_value(value=value)
        parsed_min = _parse_min_date(
            min_value=min_value,
            parsed_dates=parsed_value,
        )
        parsed_max = _parse_max_date(
            max_value=max_value,
            parsed_dates=parsed_value,
        )

        if value == "today":
            v = cast("list[date]", parsed_value)[0]
            if v < parsed_min:
                parsed_value = [parsed_min]
            if v > parsed_max:
                parsed_value = [parsed_max]

        return cls(
            value=parsed_value,
            is_range=is_range,
            min=parsed_min,
            max=parsed_max,
        )

    def __post_init__(self) -> None:
        if self.min > self.max:
            raise StreamlitAPIException(
                f"The `min_value`, set to {self.min}, shouldn't be larger "
                f"than the `max_value`, set to {self.max}."
            )

        if self.value:
            start_value = self.value[0]
            end_value = self.value[-1]

            if (start_value < self.min) or (end_value > self.max):
                raise StreamlitAPIException(
                    f"The default `value` of {self.value} "
                    f"must lie between the `min_value` of {self.min} "
                    f"and the `max_value` of {self.max}, inclusively."
                )


@dataclass
class DateTimeInputSerde:
    value: datetime | None
    min: datetime
    max: datetime

    def deserialize(self, ui_value: list[str] | None) -> datetime | None:
        if ui_value is not None and len(ui_value) > 0:
            for fmt in (_DATETIME_ISO_FORMAT, _DATETIME_LEGACY_FORMAT):
                try:
                    deserialized = _normalize_datetime_value(
                        datetime.strptime(ui_value[0], fmt)
                    )
                    break
                except ValueError:
                    continue
            else:
                # Unparseable URL query param value — revert to default.
                return self.value
            if deserialized < self.min or deserialized > self.max:
                return self.value
            return deserialized
        return self.value

    def serialize(self, v: datetime | None) -> list[str]:
        if v is None:
            return []
        return [_datetime_to_proto_string(v)]


@dataclass
class TimeInputSerde:
    value: time | None
    step: int = 900

    def deserialize(self, ui_value: str | None) -> time | None:
        if ui_value is None:
            return self.value
        try:
            # TODO(query-params): URL values that don't align to the step
            # (e.g., ?time=14:37 with step=900) are accepted as-is.
            # Consider snapping to the nearest valid step for consistency
            # with the UI. See also SliderSerde.deserialize.
            return datetime.strptime(ui_value, "%H:%M").time()
        except ValueError:
            # Unparseable URL query param value — revert to default.
            return self.value

    def serialize(self, v: datetime | time | None) -> str | None:
        if v is None:
            return None
        if isinstance(v, datetime):
            v = v.time()
        return time.strftime(v, "%H:%M")


def _to_date(v: date) -> date:
    """Convert datetime to date for comparison.

    st.date_input can receive datetime values from session_state. Since datetime
    is a subclass of date, isinstance(v, date) returns True, but datetime and date
    objects cannot be directly compared with < or >. This helper normalizes the
    value for safe comparison with date bounds.
    """
    return v.date() if isinstance(v, datetime) else v


def _validate_date_value(
    current_value: DateWidgetReturn,
    parsed_values: _DateInputValues,
    has_explicit_bounds: bool,
) -> tuple[DateWidgetReturn, bool]:
    """Validate current value against min/max bounds and reset if needed.

    Only validates when has_explicit_bounds is True (user provided min_value or max_value).
    This avoids incorrectly resetting values against computed default bounds.

    Parameters
    ----------
    current_value : DateWidgetReturn
        The current value of the date input widget. Can be a single date, a tuple of
        dates (for range mode), or None.
    parsed_values : _DateInputValues
        Parsed configuration containing min, max, default value, and whether the widget
        is in range mode.
    has_explicit_bounds : bool
        Whether the user explicitly provided min_value or max_value. If False, validation
        is skipped to avoid resetting against computed default bounds.

    Returns
    -------
    tuple[DateWidgetReturn, bool]
        A tuple of (validated_value, was_reset) where validated_value is either the
        original value (if valid) or the default value (if reset was needed), and
        was_reset indicates whether a reset occurred.
    """
    value_needs_reset = False

    if current_value is None or not has_explicit_bounds:
        return current_value, value_needs_reset

    # For range inputs, current_value is a tuple; for single inputs, it's a date
    if (
        parsed_values.is_range
        and isinstance(current_value, tuple)
        and len(current_value) > 0
    ):
        # For range mode, check if any date in the tuple is outside bounds.
        # Cast to tuple[date, ...] to satisfy the type checker after the length check.
        non_empty_value = cast("tuple[date, ...]", current_value)
        start_date = non_empty_value[0]
        end_date = non_empty_value[-1] if len(non_empty_value) > 1 else start_date
        if (
            _to_date(start_date) < parsed_values.min
            or _to_date(end_date) > parsed_values.max
        ):
            value_needs_reset = True
    elif not parsed_values.is_range and isinstance(current_value, date):
        # For single date mode. Use _to_date to handle datetime values from session_state.
        if (
            _to_date(current_value) < parsed_values.min
            or _to_date(current_value) > parsed_values.max
        ):
            value_needs_reset = True
    else:
        # Type mismatch: widget mode doesn't match current value type (e.g., range mode
        # with a single date value or single mode with a tuple). Reset to match the mode.
        value_needs_reset = True

    if not value_needs_reset:
        return current_value, value_needs_reset

    # Reset to the default value from parsed_values
    if parsed_values.value is None or len(parsed_values.value) == 0:
        return (() if parsed_values.is_range else None), True
    if not parsed_values.is_range:
        return parsed_values.value[0], True
    return cast("DateWidgetReturn", tuple(parsed_values.value)), True


def _validate_datetime_value(
    current_value: datetime | None,
    parsed_values: _DateTimeInputValues,
    has_explicit_bounds: bool,
) -> tuple[datetime | None, bool]:
    """Validate current datetime value against min/max bounds and determine if reset is needed.

    Only validates when has_explicit_bounds is True (user provided min_value or max_value).
    This avoids incorrectly determining if reset is needed against computed default bounds.

    Parameters
    ----------
    current_value : datetime | None
        The current value of the datetime input widget.
    parsed_values : _DateTimeInputValues
        Parsed configuration containing min, max, and default value.
    has_explicit_bounds : bool
        Whether the user explicitly provided min_value or max_value. If False, validation
        is skipped to avoid resetting against computed default bounds.

    Returns
    -------
    tuple[datetime | None, bool]
        A tuple of (validated_value, was_reset) where validated_value is either the
        original value (if valid) or the default value (if reset was needed), and
        was_reset indicates whether a reset occurred.
    """
    if current_value is None or not has_explicit_bounds:
        return current_value, False

    if current_value < parsed_values.min or current_value > parsed_values.max:
        return parsed_values.value, True

    return current_value, False


@dataclass
class DateInputSerde:
    value: _DateInputValues

    @staticmethod
    def _parse_date(value: str) -> date:
        """Parse a date string in ISO (YYYY-MM-DD) or legacy (YYYY/MM/DD) format."""
        for fmt in ("%Y-%m-%d", "%Y/%m/%d"):
            try:
                return datetime.strptime(value, fmt).date()
            except ValueError:  # noqa: PERF203
                continue
        raise ValueError(f"Unable to parse date: {value}")

    def deserialize(self, ui_value: Any) -> DateWidgetReturn:
        return_value: Sequence[date] | None
        if ui_value is not None:
            try:
                return_value = tuple(self._parse_date(v) for v in ui_value)
            except ValueError:
                # Invalid URL query param value (e.g. "not-a-date") — revert to default.
                return_value = self.value.value
            else:
                # Reject out-of-range dates from URL query params — revert to default.
                # Matches SliderSerde.deserialize which validates bounds in the
                # deserializer so _seed_widget_from_url can detect default equality
                # and clear the URL param via _clear_url_param.
                if return_value and any(
                    d < self.value.min or d > self.value.max for d in return_value
                ):
                    return_value = self.value.value
        else:
            return_value = self.value.value

        if return_value is None or len(return_value) == 0:
            return () if self.value.is_range else None

        if not self.value.is_range:
            return return_value[0]
        return cast("DateWidgetReturn", tuple(return_value))

    def serialize(self, v: DateWidgetReturn) -> list[str]:
        if v is None:
            return []

        to_serialize = list(v) if isinstance(v, Sequence) else [v]
        return [date.strftime(v, "%Y-%m-%d") for v in to_serialize]


class TimeWidgetsMixin:
    @overload
    def time_input(
        self,
        label: str,
        value: TimeValue = "now",
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> time:
        pass

    @overload
    def time_input(
        self,
        label: str,
        value: None = None,
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> time | None:
        pass

    @gather_metrics("time_input")
    def time_input(
        self,
        label: str,
        value: TimeValue | None = "now",
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> time | None:
        r"""Display a time input widget.

        Parameters
        ----------
        label : str
            A short label explaining to the user what this time input is for.
            The label can optionally contain GitHub-flavored Markdown of the
            following types: Bold, Italics, Strikethroughs, Inline Code, Links,
            and Images. Images display like icons, with a max height equal to
            the font height.

            Unsupported Markdown elements are unwrapped so only their children
            (text contents) render. Common block-level Markdown (headings,
            lists, blockquotes) is automatically escaped and displays as
            literal text in labels.

            See the ``body`` parameter of |st.markdown|_ for additional,
            supported Markdown directives.

            For accessibility reasons, you should never set an empty label, but
            you can hide it with ``label_visibility`` if needed. In the future,
            we may disallow empty labels by raising an exception.

            .. |st.markdown| replace:: ``st.markdown``
            .. _st.markdown: https://docs.streamlit.io/develop/api-reference/text/st.markdown

        value : "now", datetime.time, datetime.datetime, str, or None
            The value of this widget when it first renders. This can be one of
            the following:

            - ``"now"`` (default): The widget initializes with the current time.
            - A ``datetime.time`` or ``datetime.datetime`` object: The widget
              initializes with the given time, ignoring any date if included.
            - An ISO-formatted time (hh:mm[:ss.sss]) or datetime
              (YYYY-MM-DD hh:mm[:ss]) string: The widget initializes with the
              given time, ignoring any date if included.
            - ``None``: The widget initializes with no time and returns
              ``None`` until the user selects a time.

        key : str, int, or None
            An optional string or integer to use as the unique key for
            the widget. If this is ``None`` (default), a key will be
            generated for the widget based on the values of the other
            parameters. No two widgets may have the same key. Assigning
            a key stabilizes the widget's identity and preserves its
            state across reruns even when other parameters change.

            .. note::
               Changing ``step`` resets the widget even when a key is
               provided, because it constrains valid values.

            A key lets you read or update the widget's value via
            ``st.session_state[key]``. For more details, see `Widget
            behavior <https://docs.streamlit.io/develop/concepts/architecture/widget-behavior>`_.

            Additionally, if ``key`` is provided, it will be used as a
            CSS class name prefixed with ``st-key-``.

        help : str or None
            A tooltip that gets displayed next to the widget label. Streamlit
            only displays the tooltip when ``label_visibility="visible"``. If
            this is ``None`` (default), no tooltip is displayed.

            The tooltip can optionally contain GitHub-flavored Markdown,
            including the Markdown directives described in the ``body``
            parameter of ``st.markdown``.

        on_change : callable
            An optional callback invoked when this time_input's value changes.

        args : list or tuple
            An optional list or tuple of args to pass to the callback.

        kwargs : dict
            An optional dict of kwargs to pass to the callback.

        disabled : bool
            An optional boolean that disables the time input if set to
            ``True``. The default is ``False``.

        label_visibility : "visible", "hidden", or "collapsed"
            The visibility of the label. The default is ``"visible"``. If this
            is ``"hidden"``, Streamlit displays an empty spacer instead of the
            label, which can help keep the widget aligned with other widgets.
            If this is ``"collapsed"``, Streamlit displays no label or spacer.

        step : int or timedelta
            The stepping interval in seconds. This defaults to ``900`` (15
            minutes). You can also pass a ``datetime.timedelta`` object. The
            value must be between 60 seconds and 23 hours.

        width : "stretch" or int
            The width of the time input widget. This can be one of the following:

            - ``"stretch"`` (default): The width of the widget matches the
              width of the parent container.
            - An integer specifying the width in pixels: The widget has a
              fixed width. If the specified width is greater than the width of
              the parent container, the width of the widget matches the width
              of the parent container.

        bind : "query-params" or None
            Binding mode for syncing the widget's value with a URL query
            parameter. If this is ``None`` (default), the widget's value
            is not synced to the URL. When this is set to
            ``"query-params"``, changes to the widget update the URL, and
            the widget can be initialized or updated through a query
            parameter in the URL. This requires ``key`` to be set. The
            key is used as the query parameter name.

            When the widget's value equals its default, the query
            parameter is removed from the URL to keep it clean. A bound
            query parameter can't be set or deleted through
            ``st.query_params``; it can only be programmatically changed
            through ``st.session_state``.

            Times use HH:MM format in the URL. Invalid query parameter
            values are ignored and removed from the URL. If ``value``
            is ``None``, an empty query parameter (e.g., ``?my_key=``)
            clears the widget.

        Returns
        -------
        datetime.time or None
            The current value of the time input widget or ``None`` if no time has been
            selected.

        Examples
        --------
        **Example 1: Basic usage**

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> t = st.time_input("Set an alarm for", datetime.time(8, 45))
        >>> st.write("Alarm is set for", t)

        .. output::
           https://doc-time-input.streamlit.app/
           height: 260px

        **Example 2: Empty initial value**

        To initialize an empty time input, use ``None`` as the value:

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> t = st.time_input("Set an alarm for", value=None)
        >>> st.write("Alarm is set for", t)

        .. output::
           https://doc-time-input-empty.streamlit.app/
           height: 260px

        """
        ctx = get_script_run_ctx()
        return self._time_input(
            label=label,
            value=value,
            key=key,
            help=help,
            on_change=on_change,
            args=args,
            kwargs=kwargs,
            disabled=disabled,
            label_visibility=label_visibility,
            step=step,
            width=width,
            bind=bind,
            ctx=ctx,
        )

    def _time_input(
        self,
        label: str,
        value: TimeValue | None = "now",
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
        ctx: ScriptRunContext | None = None,
    ) -> time | None:
        key = to_key(key)

        check_widget_policies(
            self.dg,
            key,
            on_change,
            default_value=value if value != "now" else None,
        )
        maybe_raise_label_warnings(label, label_visibility)

        parsed_time: time | None
        parsed_time = None if value is None else _convert_timelike_to_time(value)

        element_id = compute_and_register_element_id(
            "time_input",
            user_key=key,
            # Ensure stable ID when key is provided; only whitelist step since it
            # affects the selection granularity and available options.
            key_as_main_identity={"step"},
            dg=self.dg,
            label=label,
            value=parsed_time if isinstance(value, (datetime, time)) else value,
            help=help,
            step=step,
            width=width,
        )
        del value

        session_state = get_session_state().filtered_state
        if key is not None and key in session_state and session_state[key] is None:
            parsed_time = None

        time_input_proto = TimeInputProto()
        time_input_proto.id = element_id
        time_input_proto.label = label
        if parsed_time is not None:
            time_input_proto.default = time.strftime(parsed_time, "%H:%M")
        time_input_proto.form_id = current_form_id(self.dg)
        if not isinstance(step, (int, timedelta)):
            raise StreamlitAPIException(
                f"`step` can only be `int` or `timedelta` but {type(step)} is provided."
            )
        if isinstance(step, timedelta):
            step = step.seconds
        if step < 60 or step > timedelta(hours=23).seconds:
            raise StreamlitAPIException(
                f"`step` must be between 60 seconds and 23 hours but is currently set to {step} seconds."
            )
        time_input_proto.step = step
        time_input_proto.disabled = disabled
        time_input_proto.label_visibility.value = get_label_visibility_proto_value(
            label_visibility
        )

        if help is not None:
            time_input_proto.help = dedent(help)

        if bind == "query-params" and key is not None:
            time_input_proto.query_param_key = str(key)

        serde = TimeInputSerde(parsed_time, step=step)
        widget_state = register_widget(
            time_input_proto.id,
            on_change_handler=on_change,
            args=args,
            kwargs=kwargs,
            deserializer=serde.deserialize,
            serializer=serde.serialize,
            ctx=ctx,
            value_type="string_value",
            bind=bind,
            clearable=(parsed_time is None),
        )

        if widget_state.value_changed:
            if (serialized_value := serde.serialize(widget_state.value)) is not None:
                time_input_proto.value = serialized_value
            time_input_proto.set_value = True

        layout_config = create_layout_config(width=width)

        self.dg._enqueue("time_input", time_input_proto, layout_config=layout_config)
        return widget_state.value

    @overload
    def datetime_input(
        self,
        label: str,
        value: None,
        min_value: DateTimeValue = None,
        max_value: DateTimeValue = None,
        *,  # keyword-only arguments:
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        format: str = "YYYY/MM/DD",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> datetime | None: ...

    @overload
    def datetime_input(
        self,
        label: str,
        value: DateTimeScalarValue = "now",
        min_value: DateTimeValue = None,
        max_value: DateTimeValue = None,
        *,  # keyword-only arguments:
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        format: str = "YYYY/MM/DD",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> datetime: ...

    @gather_metrics("datetime_input")
    def datetime_input(
        self,
        label: str,
        value: DateTimeValue = "now",
        min_value: DateTimeValue = None,
        max_value: DateTimeValue = None,
        *,  # keyword-only arguments:
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        format: str = "YYYY/MM/DD",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> datetime | None:
        r"""Display a date and time input widget.

        Parameters
        ----------
        label : str
            A short label explaining to the user what this datetime input is for.
            The label can optionally contain GitHub-flavored Markdown of the
            following types: Bold, Italics, Strikethroughs, Inline Code, Links,
            and Images. Images display like icons, with a max height equal to
            the font height.

            Unsupported Markdown elements are unwrapped so only their children
            (text contents) render. Common block-level Markdown (headings,
            lists, blockquotes) is automatically escaped and displays as
            literal text in labels.

            See the ``body`` parameter of |st.markdown|_ for additional,
            supported Markdown directives.

            For accessibility reasons, you should never set an empty label, but
            you can hide it with ``label_visibility`` if needed. In the future,
            we may disallow empty labels by raising an exception.

            .. |st.markdown| replace:: ``st.markdown``
            .. _st.markdown: https://docs.streamlit.io/develop/api-reference/text/st.markdown

        value : "now", datetime.datetime, datetime.date, datetime.time, str, or None
            The value of this widget when it first renders. This can be one of
            the following:

            - ``"now"`` (default): The widget initializes with the current date and time.
            - A ``datetime.datetime`` object: The widget initializes with the given
              datetime, stripping any timezone information.
            - A ``datetime.date`` object: The widget initializes with the given date
              at 00:00.
            - A ``datetime.time`` object: The widget initializes with today's date
              and the provided time.
            - An ISO-formatted datetime (YYYY-MM-DD hh:mm[:ss]) or date/time
              string: The widget initializes with the parsed value.
            - ``None``: The widget initializes with no value and returns ``None``
              until the user selects a datetime.

        min_value : "now", datetime.datetime, datetime.date, datetime.time, str, or None
            The minimum selectable datetime. This can be any of the datetime
            types accepted by ``value``.

            If this is ``None`` (default), the minimum selectable datetime is
            ten years before the initial value. If no initial value is set, the
            minimum selectable datetime is ten years before today at 00:00.

        max_value : "now", datetime.datetime, datetime.date, datetime.time, str, or None
            The maximum selectable datetime. This can be any of the datetime
            types accepted by ``value``.

            If this is ``None`` (default), the maximum selectable datetime is
            ten years after the initial value. If no initial value is set, the
            maximum selectable datetime is ten years after today at 23:59.

        key : str, int, or None
            An optional string or integer to use as the unique key for
            the widget. If this is ``None`` (default), a key will be
            generated for the widget based on the values of the other
            parameters. No two widgets may have the same key. Assigning
            a key stabilizes the widget's identity and preserves its
            state across reruns even when other parameters change.

            .. note::
               Changing ``format`` or ``step`` resets the widget even
               when a key is provided.

            A key lets you read or update the widget's value via
            ``st.session_state[key]``. For more details, see `Widget
            behavior <https://docs.streamlit.io/develop/concepts/architecture/widget-behavior>`_.

            Additionally, if ``key`` is provided, it will be used as a
            CSS class name prefixed with ``st-key-``.

        help : str or None
            A tooltip that gets displayed next to the widget label. Streamlit
            only displays the tooltip when ``label_visibility="visible"``. If
            this is ``None`` (default), no tooltip is displayed.

            The tooltip can optionally contain GitHub-flavored Markdown,
            including the Markdown directives described in the ``body``
            parameter of ``st.markdown``.

        on_change : callable
            An optional callback invoked when this datetime_input's value changes.

        args : list or tuple
            An optional list or tuple of args to pass to the callback.

        kwargs : dict
            An optional dict of kwargs to pass to the callback.

        format : str
            A format string controlling how the interface displays dates.
            Supports ``"YYYY/MM/DD"`` (default), ``"DD/MM/YYYY"``, or ``"MM/DD/YYYY"``.
            You may also use a period (.) or hyphen (-) as separators. This
            doesn't affect the time format.

        step : int or timedelta
            The stepping interval in seconds. This defaults to ``900`` (15
            minutes). You can also pass a ``datetime.timedelta`` object. The
            value must be between 60 seconds and 23 hours.

        disabled : bool
            An optional boolean that disables the widget if set to ``True``.
            The default is ``False``.

        label_visibility : "visible", "hidden", or "collapsed"
            The visibility of the label. The default is ``"visible"``. If this
            is ``"hidden"``, Streamlit displays an empty spacer instead of the
            label, which can help keep the widget aligned with other widgets.
            If this is ``"collapsed"``, Streamlit displays no label or spacer.

        width : "stretch" or int
            The width of the widget. This can be one of the following:

            - ``"stretch"`` (default): The width of the widget matches the width
              of the parent container.
            - An integer specifying the width in pixels: The widget has a fixed
              width. If the specified width is greater than the width of the
              parent container, the widget matches the container width.

        bind : "query-params" or None
            Binding mode for syncing the widget's value with a URL query
            parameter. If this is ``None`` (default), the widget's value
            is not synced to the URL. When this is set to
            ``"query-params"``, changes to the widget update the URL, and
            the widget can be initialized or updated through a query
            parameter in the URL. This requires ``key`` to be set. The
            key is used as the query parameter name.

            When the widget's value equals its default, the query
            parameter is removed from the URL to keep it clean. A bound
            query parameter can't be set or deleted through
            ``st.query_params``; it can only be programmatically changed
            through ``st.session_state``.

            Datetimes use ISO 8601 format (YYYY-MM-DDThh:mm) in the URL.
            Invalid query parameter values are ignored and removed from
            the URL. If ``value`` is ``None``, an empty query parameter
            (e.g., ``?my_key=``) clears the widget.

        Returns
        -------
        datetime.datetime or None
            The current value of the datetime input widget (without timezone)
            or ``None`` if no value has been selected.

        Examples
        --------
        **Example 1: Basic usage**

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> event_time = st.datetime_input(
        ...     "Schedule your event",
        ...     datetime.datetime(2025, 11, 19, 16, 45),
        ... )
        >>> st.write("Event scheduled for", event_time)

        .. output::
           https://doc-datetime-input.streamlit.app/
           height: 500px

        **Example 2: Empty initial value**

        To initialize an empty datetime input, use ``None`` as the value:

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> event_time = st.datetime_input("Schedule your event", value=None)
        >>> st.write("Event scheduled for", event_time)

        .. output::
           https://doc-datetime-input-empty.streamlit.app/
           height: 500px

        """
        ctx = get_script_run_ctx()
        return self._datetime_input(
            label=label,
            value=value,
            min_value=min_value,
            max_value=max_value,
            key=key,
            help=help,
            on_change=on_change,
            args=args,
            kwargs=kwargs,
            format=format,
            step=step,
            disabled=disabled,
            label_visibility=label_visibility,
            width=width,
            bind=bind,
            ctx=ctx,
        )

    def _datetime_input(
        self,
        label: str,
        value: DateTimeValue = "now",
        min_value: DateTimeValue = None,
        max_value: DateTimeValue = None,
        *,  # keyword-only arguments:
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        format: str = "YYYY/MM/DD",
        step: int | timedelta = timedelta(minutes=DEFAULT_STEP_MINUTES),
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
        ctx: ScriptRunContext | None = None,
    ) -> datetime | None:
        key = to_key(key)

        check_widget_policies(
            self.dg,
            key,
            on_change,
            default_value=value if value != "now" else None,
        )
        maybe_raise_label_warnings(label, label_visibility)

        datetime_values = _DateTimeInputValues.from_raw_values(
            value=value,
            min_value=min_value,
            max_value=max_value,
        )

        default_value = datetime_values.value
        min_value_proto = _datetime_to_proto_string(datetime_values.min)
        max_value_proto = _datetime_to_proto_string(datetime_values.max)

        if isinstance(value, (datetime, date, time)):
            value_for_id: Any = (
                None
                if default_value is None
                else _datetime_to_proto_string(default_value)
            )
        else:
            value_for_id = value

        element_id = compute_and_register_element_id(
            "date_time_input",
            user_key=key,
            # Format is whitelisted because of a bug in the BaseWeb date input component.
            # Step is whitelisted because it invalidates the current selection.
            # We might be able to unlock this as a follow-up.
            key_as_main_identity={"format", "step"},
            dg=self.dg,
            label=label,
            value=value_for_id,
            min_value=min_value_proto,
            max_value=max_value_proto,
            help=help,
            format=format,
            step=step,
            width=width,
        )
        # Track if user explicitly set bounds (before del)
        has_explicit_bounds = min_value is not None or max_value is not None
        del value, min_value, max_value

        if not bool(ALLOWED_DATE_FORMATS.match(format)):
            raise StreamlitAPIException(
                f"The provided format (`{format}`) is not valid. DateTimeInput format "
                "should be one of `YYYY/MM/DD`, `DD/MM/YYYY`, or `MM/DD/YYYY` "
                "and can also use a period (.) or hyphen (-) as separators."
            )

        if not isinstance(step, (int, timedelta)):
            raise StreamlitAPIException(
                f"`step` can only be `int` or `timedelta` but {type(step)} is provided."
            )
        step_seconds = (
            int(step.total_seconds()) if isinstance(step, timedelta) else step
        )
        if step_seconds < 60 or step_seconds > timedelta(hours=23).seconds:
            raise StreamlitAPIException(
                f"`step` must be between 60 seconds and 23 hours but is currently set to {step_seconds} seconds."
            )

        session_state = get_session_state().filtered_state
        default_value_for_proto = default_value
        if key is not None and key in session_state and session_state[key] is None:
            default_value_for_proto = None

        date_time_input_proto = DateTimeInputProto()
        date_time_input_proto.id = element_id
        date_time_input_proto.label = label
        if default_value_for_proto is not None:
            date_time_input_proto.default[:] = [
                _datetime_to_proto_string(default_value_for_proto)
            ]
        date_time_input_proto.min = min_value_proto
        date_time_input_proto.max = max_value_proto
        date_time_input_proto.form_id = current_form_id(self.dg)
        date_time_input_proto.step = step_seconds
        date_time_input_proto.disabled = disabled
        date_time_input_proto.label_visibility.value = get_label_visibility_proto_value(
            label_visibility
        )
        date_time_input_proto.format = format
        date_time_input_proto.is_range = False

        if help is not None:
            date_time_input_proto.help = dedent(help)

        if bind == "query-params" and key is not None:
            date_time_input_proto.query_param_key = str(key)

        serde = DateTimeInputSerde(
            value=default_value_for_proto,
            min=datetime_values.min,
            max=datetime_values.max,
        )
        widget_state = register_widget(
            date_time_input_proto.id,
            on_change_handler=on_change,
            args=args,
            kwargs=kwargs,
            deserializer=serde.deserialize,
            serializer=serde.serialize,
            ctx=ctx,
            value_type="string_array_value",
            bind=bind,
            clearable=(default_value is None),
        )

        # Validate the current value against the new min/max bounds.
        # Only validate when user explicitly provided min_value or max_value.
        current_value, value_needs_reset = _validate_datetime_value(
            widget_state.value, datetime_values, has_explicit_bounds
        )

        if value_needs_reset and key is not None:
            # Update session_state so subsequent accesses in this run
            # return the corrected value. Use reset_state_value to avoid
            # the "cannot be modified after widget instantiated" error.
            get_session_state().reset_state_value(key, current_value)

            # Clear stale URL param when an out-of-bounds URL value was reset.
            if bind == "query-params":
                with get_session_state().query_params() as qp:
                    qp.remove_param(str(key))

        if value_needs_reset or widget_state.value_changed:
            date_time_input_proto.value[:] = serde.serialize(current_value)
            date_time_input_proto.set_value = True

        layout_config = create_layout_config(width=width)

        self.dg._enqueue(
            "date_time_input", date_time_input_proto, layout_config=layout_config
        )
        return current_value

    @overload
    def date_input(
        self,
        label: str,
        value: date | datetime | str | Literal["today"] = "today",
        min_value: NullableScalarDateValue = None,
        max_value: NullableScalarDateValue = None,
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        format: str = "YYYY/MM/DD",
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> date: ...

    @overload
    def date_input(
        self,
        label: str,
        value: None,
        min_value: NullableScalarDateValue = None,
        max_value: NullableScalarDateValue = None,
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        format: str = "YYYY/MM/DD",
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> date | None: ...

    @overload
    def date_input(
        self,
        label: str,
        value: tuple[NullableScalarDateValue]
        | tuple[NullableScalarDateValue, NullableScalarDateValue]
        | list[NullableScalarDateValue],
        min_value: NullableScalarDateValue = None,
        max_value: NullableScalarDateValue = None,
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        format: str = "YYYY/MM/DD",
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> DateWidgetRangeReturn: ...

    @gather_metrics("date_input")
    def date_input(
        self,
        label: str,
        value: DateValue = "today",
        min_value: NullableScalarDateValue = None,
        max_value: NullableScalarDateValue = None,
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        format: str = "YYYY/MM/DD",
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
    ) -> DateWidgetReturn:
        r"""Display a date input widget.

        The date input widget can be configured to accept a single date or a
        date range. The first day of the week is determined from the user's
        locale in their browser.

        Parameters
        ----------
        label : str
            A short label explaining to the user what this date input is for.
            The label can optionally contain GitHub-flavored Markdown of the
            following types: Bold, Italics, Strikethroughs, Inline Code, Links,
            and Images. Images display like icons, with a max height equal to
            the font height.

            Unsupported Markdown elements are unwrapped so only their children
            (text contents) render. Common block-level Markdown (headings,
            lists, blockquotes) is automatically escaped and displays as
            literal text in labels.

            See the ``body`` parameter of |st.markdown|_ for additional,
            supported Markdown directives.

            For accessibility reasons, you should never set an empty label, but
            you can hide it with ``label_visibility`` if needed. In the future,
            we may disallow empty labels by raising an exception.

            .. |st.markdown| replace:: ``st.markdown``
            .. _st.markdown: https://docs.streamlit.io/develop/api-reference/text/st.markdown

        value : "today", datetime.date, datetime.datetime, str, list/tuple of these, or None
            The value of this widget when it first renders. This can be one of
            the following:

            - ``"today"`` (default): The widget initializes with the current date.
            - A ``datetime.date`` or ``datetime.datetime`` object: The widget
              initializes with the given date, ignoring any time if included.
            - An ISO-formatted date (YYYY-MM-DD) or datetime
              (YYYY-MM-DD hh:mm:ss) string: The widget initializes with the
              given date, ignoring any time if included.
            - A list or tuple with up to two of the above: The widget will
              initialize with the given date interval and return a tuple of the
              selected interval. You can pass an empty list to initialize the
              widget with an empty interval or a list with one value to
              initialize only the beginning date of the iterval.
            - ``None``: The widget initializes with no date and returns
              ``None`` until the user selects a date.

        min_value : "today", datetime.date, datetime.datetime, str, or None
            The minimum selectable date. This can be any of the date types
            accepted by ``value``, except list or tuple.

            If this is ``None`` (default), the minimum selectable date is ten
            years before the initial value. If the initial value is an
            interval, the minimum selectable date is ten years before the start
            date of the interval. If no initial value is set, the minimum
            selectable date is ten years before today.

        max_value : "today", datetime.date, datetime.datetime, str, or None
            The maximum selectable date. This can be any of the date types
            accepted by ``value``, except list or tuple.

            If this is ``None`` (default), the maximum selectable date is ten
            years after the initial value. If the initial value is an interval,
            the maximum selectable date is ten years after the end date of the
            interval. If no initial value is set, the maximum selectable date
            is ten years after today.

        key : str, int, or None
            An optional string or integer to use as the unique key for
            the widget. If this is ``None`` (default), a key will be
            generated for the widget based on the values of the other
            parameters. No two widgets may have the same key. Assigning
            a key stabilizes the widget's identity and preserves its
            state across reruns even when other parameters change.

            .. note::
               Changing ``format`` resets the widget even when a key is
               provided.

            A key lets you read or update the widget's value via
            ``st.session_state[key]``. For more details, see `Widget
            behavior <https://docs.streamlit.io/develop/concepts/architecture/widget-behavior>`_.

            Additionally, if ``key`` is provided, it will be used as a
            CSS class name prefixed with ``st-key-``.

        help : str or None
            A tooltip that gets displayed next to the widget label. Streamlit
            only displays the tooltip when ``label_visibility="visible"``. If
            this is ``None`` (default), no tooltip is displayed.

            The tooltip can optionally contain GitHub-flavored Markdown,
            including the Markdown directives described in the ``body``
            parameter of ``st.markdown``.

        on_change : callable
            An optional callback invoked when this date_input's value changes.

        args : list or tuple
            An optional list or tuple of args to pass to the callback.

        kwargs : dict
            An optional dict of kwargs to pass to the callback.

        format : str
            A format string controlling how the interface should display dates.
            Supports ``"YYYY/MM/DD"`` (default), ``"DD/MM/YYYY"``, or ``"MM/DD/YYYY"``.
            You may also use a period (.) or hyphen (-) as separators.

        disabled : bool
            An optional boolean that disables the date input if set to
            ``True``. The default is ``False``.

        label_visibility : "visible", "hidden", or "collapsed"
            The visibility of the label. The default is ``"visible"``. If this
            is ``"hidden"``, Streamlit displays an empty spacer instead of the
            label, which can help keep the widget aligned with other widgets.
            If this is ``"collapsed"``, Streamlit displays no label or spacer.

        width : "stretch" or int
            The width of the date input widget. This can be one of the following:

            - ``"stretch"`` (default): The width of the widget matches the
              width of the parent container.
            - An integer specifying the width in pixels: The widget has a
              fixed width. If the specified width is greater than the width of
              the parent container, the width of the widget matches the width
              of the parent container.

        bind : "query-params" or None
            Binding mode for syncing the widget's value with a URL query
            parameter. If this is ``None`` (default), the widget's value
            is not synced to the URL. When this is set to
            ``"query-params"``, changes to the widget update the URL, and
            the widget can be initialized or updated through a query
            parameter in the URL. This requires ``key`` to be set. The
            key is used as the query parameter name.

            When the widget's value equals its default, the query
            parameter is removed from the URL to keep it clean. A bound
            query parameter can't be set or deleted through
            ``st.query_params``; it can only be programmatically changed
            through ``st.session_state``.

            Dates use ISO 8601 format (YYYY-MM-DD) in the URL. Invalid
            query parameter values are ignored and removed from the URL.
            If ``value`` is ``None``, an empty query parameter (e.g.,
            ``?vacation=``) clears the widget. Date ranges use repeated
            parameters (e.g.,
            ``?vacation=2025-01-01&vacation=2025-01-31``).

        Returns
        -------
        datetime.date or a tuple with 0-2 dates or None
            The current value of the date input widget or ``None`` if no date has been
            selected.

        Examples
        --------
        **Example 1: Basic usage**

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> d = st.date_input("When's your birthday", datetime.date(2019, 7, 6))
        >>> st.write("Your birthday is:", d)

        .. output::
           https://doc-date-input.streamlit.app/
           height: 380px

        **Example 2: Date range**

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> today = datetime.datetime.now()
        >>> next_year = today.year + 1
        >>> jan_1 = datetime.date(next_year, 1, 1)
        >>> dec_31 = datetime.date(next_year, 12, 31)
        >>>
        >>> d = st.date_input(
        ...     "Select your vacation for next year",
        ...     (jan_1, datetime.date(next_year, 1, 7)),
        ...     jan_1,
        ...     dec_31,
        ...     format="MM.DD.YYYY",
        ... )
        >>> d

        .. output::
           https://doc-date-input1.streamlit.app/
           height: 380px

        **Example 3: Empty initial value**

        To initialize an empty date input, use ``None`` as the value:

        >>> import datetime
        >>> import streamlit as st
        >>>
        >>> d = st.date_input("When's your birthday", value=None)
        >>> st.write("Your birthday is:", d)

        .. output::
           https://doc-date-input-empty.streamlit.app/
           height: 380px

        """
        ctx = get_script_run_ctx()
        return self._date_input(
            label=label,
            value=value,
            min_value=min_value,
            max_value=max_value,
            key=key,
            help=help,
            on_change=on_change,
            args=args,
            kwargs=kwargs,
            disabled=disabled,
            label_visibility=label_visibility,
            format=format,
            width=width,
            bind=bind,
            ctx=ctx,
        )

    def _date_input(
        self,
        label: str,
        value: DateValue = "today",
        min_value: NullableScalarDateValue = None,
        max_value: NullableScalarDateValue = None,
        key: Key | None = None,
        help: str | None = None,
        on_change: WidgetCallback | None = None,
        args: WidgetArgs | None = None,
        kwargs: WidgetKwargs | None = None,
        *,  # keyword-only arguments:
        format: str = "YYYY/MM/DD",
        disabled: bool = False,
        label_visibility: LabelVisibility = "visible",
        width: WidthWithoutContent = "stretch",
        bind: BindOption = None,
        ctx: ScriptRunContext | None = None,
    ) -> DateWidgetReturn:
        key = to_key(key)

        check_widget_policies(
            self.dg,
            key,
            on_change,
            default_value=value if value != "today" else None,
        )
        maybe_raise_label_warnings(label, label_visibility)

        def parse_date_deterministic_for_id(v: NullableScalarDateValue) -> str | None:
            if v == "today":
                # For ID purposes, no need to parse the input string.
                return None
            if isinstance(v, str):
                # For ID purposes, no need to parse the input string.
                return v
            if isinstance(v, datetime):
                return date.strftime(v.date(), "%Y-%m-%d")
            if isinstance(v, date):
                return date.strftime(v, "%Y-%m-%d")

            return None

        parsed_min_date = parse_date_deterministic_for_id(min_value)
        parsed_max_date = parse_date_deterministic_for_id(max_value)

        parsed: str | list[str | None] | None
        if value == "today":
            parsed = None
        elif isinstance(value, Sequence):
            parsed = [parse_date_deterministic_for_id(v) for v in value]  # ty: ignore[invalid-argument-type]
        else:
            parsed = parse_date_deterministic_for_id(value)

        # TODO: this is missing the error path, integrate with the dateinputvalues parsing

        element_id = compute_and_register_element_id(
            "date_input",
            user_key=key,
            # Ensure stable ID when key is provided. Only format is whitelisted because
            # there is a bug in baseweb where changing the format dynamically leads to
            # a wrongly formatted date. min_value and max_value support dynamic changes.
            key_as_main_identity={"format"},
            dg=self.dg,
            label=label,
            value=parsed,
            min_value=parsed_min_date,
            max_value=parsed_max_date,
            help=help,
            format=format,
            width=width,
        )
        if not bool(ALLOWED_DATE_FORMATS.match(format)):
            raise StreamlitAPIException(
                f"The provided format (`{format}`) is not valid. DateInput format "
                "should be one of `YYYY/MM/DD`, `DD/MM/YYYY`, or `MM/DD/YYYY` "
                "and can also use a period (.) or hyphen (-) as separators."
            )

        parsed_values = _DateInputValues.from_raw_values(
            value=value,
            min_value=min_value,
            max_value=max_value,
        )

        # Track if user explicitly set bounds (before del)
        has_explicit_bounds = min_value is not None or max_value is not None

        if value == "today":
            # We need to know if this is a single or range date_input, but don't have
            # a default value, so we check if session_state can tell us.
            # We already calculated the id, so there is no risk of this causing
            # the id to change.

            session_state = get_session_state().filtered_state

            if key is not None and key in session_state:
                state_value = session_state[key]
                parsed_values = _DateInputValues.from_raw_values(
                    value=state_value,
                    min_value=min_value,
                    max_value=max_value,
                )

        del value, min_value, max_value

        date_input_proto = DateInputProto()
        date_input_proto.id = element_id
        date_input_proto.is_range = parsed_values.is_range
        date_input_proto.disabled = disabled
        date_input_proto.label_visibility.value = get_label_visibility_proto_value(
            label_visibility
        )
        date_input_proto.format = format
        date_input_proto.label = label
        if parsed_values.value is None:
            # An empty array represents the empty state. The reason for using an empty
            # array here is that we cannot optional keyword for repeated fields
            # in protobuf.
            date_input_proto.default[:] = []
        else:
            date_input_proto.default[:] = [
                date.strftime(v, "%Y-%m-%d") for v in parsed_values.value
            ]
        date_input_proto.min = date.strftime(parsed_values.min, "%Y-%m-%d")
        date_input_proto.max = date.strftime(parsed_values.max, "%Y-%m-%d")
        date_input_proto.form_id = current_form_id(self.dg)

        if help is not None:
            date_input_proto.help = dedent(help)

        if bind == "query-params" and key is not None:
            date_input_proto.query_param_key = str(key)

        serde = DateInputSerde(parsed_values)

        widget_state = register_widget(
            date_input_proto.id,
            on_change_handler=on_change,
            args=args,
            kwargs=kwargs,
            deserializer=serde.deserialize,
            serializer=serde.serialize,
            ctx=ctx,
            value_type="string_array_value",
            bind=bind,
            clearable=(parsed_values.value is None),
        )

        # Validate the current value against the new min/max bounds.
        # Only validate when user explicitly provided min_value or max_value.
        current_value, value_needs_reset = _validate_date_value(
            widget_state.value, parsed_values, has_explicit_bounds
        )

        # Reset if needed.
        if value_needs_reset and key is not None:
            # Update session_state so subsequent accesses in this run
            # return the corrected value. Use reset_state_value to avoid
            # the "cannot be modified after widget instantiated" error.
            get_session_state().reset_state_value(key, current_value)

            # Clear stale URL param when an out-of-bounds URL value was reset.
            if bind == "query-params":
                with get_session_state().query_params() as qp:
                    qp.remove_param(str(key))

        if value_needs_reset or widget_state.value_changed:
            date_input_proto.value[:] = serde.serialize(current_value)
            date_input_proto.set_value = True

        layout_config = create_layout_config(width=width)

        self.dg._enqueue("date_input", date_input_proto, layout_config=layout_config)
        return current_value

    @property
    def dg(self) -> DeltaGenerator:
        """Get our DeltaGenerator."""
        return cast("DeltaGenerator", self)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: streamlit/proto/Metric.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'streamlit/proto/Metric.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from streamlit.proto import LabelVisibility_pb2 as streamlit_dot_proto_dot_LabelVisibility__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cstreamlit/proto/Metric.proto\x1a%streamlit/proto/LabelVisibility.proto\"\xfa\x03\n\x06Metric\x12\r\n\x05label\x18\x01 \x01(\t\x12\x0c\n\x04\x62ody\x18\x02 \x01(\t\x12\r\n\x05\x64\x65lta\x18\x03 \x01(\t\x12*\n\tdirection\x18\x04 \x01(\x0e\x32\x17.Metric.MetricDirection\x12\"\n\x05\x63olor\x18\x05 \x01(\x0e\x32\x13.Metric.MetricColor\x12\x0c\n\x04help\x18\x06 \x01(\t\x12*\n\x10label_visibility\x18\x07 \x01(\x0b\x32\x10.LabelVisibility\x12\x13\n\x0bshow_border\x18\x08 \x01(\x08\x12\x12\n\nchart_data\x18\t \x03(\x01\x12%\n\nchart_type\x18\n \x01(\x0e\x32\x11.Metric.ChartType\x12\x0e\n\x06\x66ormat\x18\x0b \x01(\t\x12\x19\n\x11\x64\x65lta_description\x18\x0c \x01(\t\"-\n\x0fMetricDirection\x12\x08\n\x04\x44OWN\x10\x00\x12\x06\n\x02UP\x10\x01\x12\x08\n\x04NONE\x10\x02\"f\n\x0bMetricColor\x12\x07\n\x03RED\x10\x00\x12\t\n\x05GREEN\x10\x01\x12\x08\n\x04GRAY\x10\x02\x12\n\n\x06ORANGE\x10\x03\x12\n\n\x06YELLOW\x10\x04\x12\x08\n\x04\x42LUE\x10\x05\x12\n\n\x06VIOLET\x10\x06\x12\x0b\n\x07PRIMARY\x10\x07\"(\n\tChartType\x12\x08\n\x04LINE\x10\x00\x12\x07\n\x03\x42\x41R\x10\x01\x12\x08\n\x04\x41REA\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'streamlit.proto.Metric_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_METRIC']._serialized_start=72
  _globals['_METRIC']._serialized_end=578
  _globals['_METRIC_METRICDIRECTION']._serialized_start=387
  _globals['_METRIC_METRICDIRECTION']._serialized_end=432
  _globals['_METRIC_METRICCOLOR']._serialized_start=434
  _globals['_METRIC_METRICCOLOR']._serialized_end=536
  _globals['_METRIC_CHARTTYPE']._serialized_start=538
  _globals['_METRIC_CHARTTYPE']._serialized_end=578
# @@protoc_insertion_point(module_scope)



# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Literal, TypeAlias, cast

from streamlit.elements.lib.layout_utils import create_layout_config
from streamlit.errors import StreamlitAPIException
from streamlit.proto.Heading_pb2 import Heading as HeadingProto
from streamlit.runtime.metrics_util import gather_metrics
from streamlit.string_util import clean_text

if TYPE_CHECKING:
    from streamlit.delta_generator import DeltaGenerator
    from streamlit.elements.lib.layout_utils import TextAlignment, Width
    from streamlit.type_util import SupportsStr


class HeadingProtoTag(Enum):
    TITLE_TAG = "h1"
    HEADER_TAG = "h2"
    SUBHEADER_TAG = "h3"


Anchor: TypeAlias = str | Literal[False] | None
Divider: TypeAlias = bool | str | None


class HeadingMixin:
    @gather_metrics("header")
    def header(
        self,
        body: SupportsStr,
        anchor: Anchor = None,
        *,  # keyword-only arguments:
        help: str | None = None,
        divider: Divider = False,
        width: Width = "stretch",
        text_alignment: TextAlignment = "left",
    ) -> DeltaGenerator:
        """Display text in header formatting.

        Parameters
        ----------
        body : str
            The text to display as GitHub-flavored Markdown. Syntax
            information can be found at: https://github.github.com/gfm.

            See the ``body`` parameter of |st.markdown|_ for additional,
            supported Markdown directives.

            .. |st.markdown| replace:: ``st.markdown``
            .. _st.markdown: https://docs.streamlit.io/develop/api-reference/text/st.markdown

        anchor : str or False
            The anchor name of the header that can be accessed with #anchor
            in the URL. If omitted, it generates an anchor using the body.
            If False, the anchor is not shown in the UI.

        help : str or None
            A tooltip that gets displayed next to the header. If this is
            ``None`` (default), no tooltip is displayed.

            The tooltip can optionally contain GitHub-flavored Markdown,
            including the Markdown directives described in the ``body``
            parameter of ``st.markdown``.

        divider : bool, "blue", "green", "orange", "red", "violet", "yellow", "gray"/"grey", or "rainbow"
            Shows a colored divider below the header. If this is ``True``,
            successive headers will cycle through divider colors, except gray
            and rainbow. That is, the first header will have a blue line, the
            second header will have a green line, and so on. If this is a
            string, the color can be set to one of the following: blue, green,
            orange, red, violet, yellow, gray/grey, or rainbow.

        width : "stretch", "content", or int
            The width of the header element. This can be one of the following:

            - ``"stretch"`` (default): The width of the element matches the
              width of the parent container.
            - ``"content"``: The width of the element matches the width of its
              content, but doesn't exceed the width of the parent container.
            - An integer specifying the width in pixels: The element has a
              fixed width. If the specified width is greater than the width of
              the parent container, the width of the element matches the width
              of the parent container.

        text_alignment : "left", "center", "right", or "justify"
            The horizontal alignment of the text within the element. This can
            be one of the following:

            - ``"left"`` (default): Text is aligned to the left edge.
            - ``"center"``: Text is centered.
            - ``"right"``: Text is aligned to the right edge.
            - ``"justify"``: Text is justified (stretched to fill the available
              width with the last line left-aligned).

            .. note::
                For text alignment to have a visible effect, the element's
                width must be wider than its content. If you use
                ``width="content"`` with short text, the alignment may not be
                noticeable.

        Examples
        --------
        >>> import streamlit as st
        >>>
        >>> st.header("_Streamlit_ is :blue[cool] :sunglasses:")
        >>> st.header("This is a header with a divider", divider="gray")
        >>> st.header("These headers have rotating dividers", divider=True)
        >>> st.header("One", divider=True)
        >>> st.header("Two", divider=True)
        >>> st.header("Three", divider=True)
        >>> st.header("Four", divider=True)

        .. output::
           https://doc-header.streamlit.app/
           height: 600px

        """
        layout_config = create_layout_config(
            width=width,
            text_alignment=text_alignment,
            allow_content_width=True,
        )

        return self.dg._enqueue(
            "heading",
            HeadingMixin._create_heading_proto(
                tag=HeadingProtoTag.HEADER_TAG,
                body=body,
                anchor=anchor,
                help=help,
                divider=divider,
            ),
            layout_config=layout_config,
        )

    @gather_metrics("subheader")
    def subheader(
        self,
        body: SupportsStr,
        anchor: Anchor = None,
        *,  # keyword-only arguments:
        help: str | None = None,
        divider: Divider = False,
        width: Width = "stretch",
        text_alignment: TextAlignment = "left",
    ) -> DeltaGenerator:
        """Display text in subheader formatting.

        Parameters
        ----------
        body : str
            The text to display as GitHub-flavored Markdown. Syntax
            information can be found at: https://github.github.com/gfm.

            See the ``body`` parameter of |st.markdown|_ for additional,
            supported Markdown directives.

            .. |st.markdown| replace:: ``st.markdown``
            .. _st.markdown: https://docs.streamlit.io/develop/api-reference/text/st.markdown

        anchor : str or False
            The anchor name of the header that can be accessed with #anchor
            in the URL. If omitted, it generates an anchor using the body.
            If False, the anchor is not shown in the UI.

        help : str or None
            A tooltip that gets displayed next to the subheader. If this is
            ``None`` (default), no tooltip is displayed.

            The tooltip can optionally contain GitHub-flavored Markdown,
            including the Markdown directives described in the ``body``
            parameter of ``st.markdown``.

        divider : bool, "blue", "green", "orange", "red", "violet", "yellow", "gray"/"grey", or "rainbow"
            Shows a colored divider below the header. If this is ``True``,
            successive headers will cycle through divider colors, except gray
            and rainbow. That is, the first header will have a blue line, the
            second header will have a green line, and so on. If this is a
            string, the color can be set to one of the following: blue, green,
            orange, red, violet, yellow, gray/grey, or rainbow.

        width : "stretch", "content", or int
            The width of the subheader element. This can be one of the following:

            - ``"stretch"`` (default): The width of the element matches the
              width of the parent container.
            - ``"content"``: The width of the element matches the width of its
              content, but doesn't exceed the width of the parent container.
            - An integer specifying the width in pixels: The element has a
              fixed width. If the specified width is greater than the width of
              the parent container, the width of the element matches the width
              of the parent container.

        text_alignment : "left", "center", "right", or "justify"
            The horizontal alignment of the text within the element. This can
            be one of the following:

            - ``"left"`` (default): Text is aligned to the left edge.
            - ``"center"``: Text is centered.
            - ``"right"``: Text is aligned to the right edge.
            - ``"justify"``: Text is justified (stretched to fill the available
              width with the last line left-aligned).

            .. note::
                For text alignment to have a visible effect, the element's
                width must be wider than its content. If you use
                ``width="content"`` with short text, the alignment may not be
                noticeable.

        Examples
        --------
        >>> import streamlit as st
        >>>
        >>> st.subheader("_Streamlit_ is :blue[cool] :sunglasses:")
        >>> st.subheader("This is a subheader with a divider", divider="gray")
        >>> st.subheader("These subheaders have rotating dividers", divider=True)
        >>> st.subheader("One", divider=True)
        >>> st.subheader("Two", divider=True)
        >>> st.subheader("Three", divider=True)
        >>> st.subheader("Four", divider=True)

        .. output::
           https://doc-subheader.streamlit.app/
           height: 500px

        """
        layout_config = create_layout_config(
            width=width,
            text_alignment=text_alignment,
            allow_content_width=True,
        )

        return self.dg._enqueue(
            "heading",
            HeadingMixin._create_heading_proto(
                tag=HeadingProtoTag.SUBHEADER_TAG,
                body=body,
                anchor=anchor,
                help=help,
                divider=divider,
            ),
            layout_config=layout_config,
        )

    @gather_metrics("title")
    def title(
        self,
        body: SupportsStr,
        anchor: Anchor = None,
        *,  # keyword-only arguments:
        help: str | None = None,
        width: Width = "stretch",
        text_alignment: TextAlignment = "left",
    ) -> DeltaGenerator:
        """Display text in title formatting.

        Each document should have a single `st.title()`, although this is not
        enforced.

        Parameters
        ----------
        body : str
            The text to display as GitHub-flavored Markdown. Syntax
            information can be found at: https://github.github.com/gfm.

            See the ``body`` parameter of |st.markdown|_ for additional,
            supported Markdown directives.

            .. |st.markdown| replace:: ``st.markdown``
            .. _st.markdown: https://docs.streamlit.io/develop/api-reference/text/st.markdown

        anchor : str or False
            The anchor name of the header that can be accessed with #anchor
            in the URL. If omitted, it generates an anchor using the body.
            If False, the anchor is not shown in the UI.

        help : str or None
            A tooltip that gets displayed next to the title. If this is
            ``None`` (default), no tooltip is displayed.

            The tooltip can optionally contain GitHub-flavored Markdown,
            including the Markdown directives described in the ``body``
            parameter of ``st.markdown``.

        width : "stretch", "content", or int
            The width of the title element. This can be one of the following:

            - ``"stretch"`` (default): The width of the element matches the
              width of the parent container.
            - ``"content"``: The width of the element matches the width of its
              content, but doesn't exceed the width of the parent container.
            - An integer specifying the width in pixels: The element has a
              fixed width. If the specified width is greater than the width of
              the parent container, the width of the element matches the width
              of the parent container.

        text_alignment : "left", "center", "right", or "justify"
            The horizontal alignment of the text within the element. This can
            be one of the following:

            - ``"left"`` (default): Text is aligned to the left edge.
            - ``"center"``: Text is centered.
            - ``"right"``: Text is aligned to the right edge.
            - ``"justify"``: Text is justified (stretched to fill the available
              width with the last line left-aligned).

            .. note::
                For text alignment to have a visible effect, the element's
                width must be wider than its content. If you use
                ``width="content"`` with short text, the alignment may not be
                noticeable.

        Examples
        --------
        >>> import streamlit as st
        >>>
        >>> st.title("This is a title")
        >>> st.title("_Streamlit_ is :blue[cool] :sunglasses:")

        .. output::
           https://doc-title.streamlit.app/
           height: 220px

        """
        layout_config = create_layout_config(
            width=width,
            text_alignment=text_alignment,
            allow_content_width=True,
        )

        return self.dg._enqueue(
            "heading",
            HeadingMixin._create_heading_proto(
                tag=HeadingProtoTag.TITLE_TAG,
                body=body,
                anchor=anchor,
                help=help,
            ),
            layout_config=layout_config,
        )

    @property
    def dg(self) -> DeltaGenerator:
        """Get our DeltaGenerator."""
        return cast("DeltaGenerator", self)

    @staticmethod
    def _handle_divider_color(divider: Divider) -> str:
        if divider is True:
            return "auto"
        valid_colors = [
            "red",
            "orange",
            "yellow",
            "blue",
            "green",
            "violet",
            "gray",
            "grey",
            "rainbow",
        ]
        if divider in valid_colors:
            return cast("str", divider)  # ty: ignore[redundant-cast]
        raise StreamlitAPIException(
            f"Divider parameter has invalid value: `{divider}`. Please choose from: {', '.join(valid_colors)}."
        )

    @staticmethod
    def _create_heading_proto(
        tag: HeadingProtoTag,
        body: SupportsStr,
        anchor: Anchor = None,
        help: str | None = None,
        divider: Divider = False,
    ) -> HeadingProto:
        proto = HeadingProto()
        proto.tag = tag.value
        proto.body = clean_text(body)
        if divider:
            proto.divider = HeadingMixin._handle_divider_color(divider)
        if anchor is not None:
            if anchor is False:
                proto.hide_anchor = True
            elif isinstance(anchor, str):
                proto.anchor = anchor
            elif anchor is True:  # type: ignore
                raise StreamlitAPIException(
                    f"Anchor parameter has invalid value: {anchor}. "
                    "Supported values: None, any string or False"
                )
            else:
                raise StreamlitAPIException(
                    f"Anchor parameter has invalid type: {type(anchor).__name__}. "
                    "Supported values: None, any string or False"
                )

        if help:
            proto.help = help

        return proto



# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The main purpose of this module (right now at least) is to avoid a dependency
cycle between streamlit.delta_generator and some elements.
"""

from __future__ import annotations

from contextvars import ContextVar, Token
from typing import TYPE_CHECKING, Generic, TypeVar

from streamlit.proto.RootContainer_pb2 import RootContainer as _RootContainer

if TYPE_CHECKING:
    from collections.abc import Callable

    from streamlit.delta_generator import DeltaGenerator
    from streamlit.elements.lib.dialog import Dialog
    from streamlit.elements.lib.mutable_expander_container import ExpanderContainer
    from streamlit.elements.lib.mutable_popover_container import PopoverContainer
    from streamlit.elements.lib.mutable_status_container import StatusContainer
    from streamlit.elements.lib.mutable_tab_container import TabContainer


class DeltaGeneratorSingleton:
    """Used to initialize the DeltaGenerator classes and store them as singletons.
    This module allows us to avoid circular imports between DeltaGenerator and elements,
    because elements can import this singleton module instead of DeltaGenerator directly.
    """

    _instance: DeltaGeneratorSingleton | None = None

    @classmethod
    def instance(cls) -> DeltaGeneratorSingleton:
        """Return the singleton DeltaGeneratorSingleton instance. Raise an Error if the
        DeltaGeneratorSingleton hasn't been created yet.
        """
        if cls._instance is None:
            raise RuntimeError("DeltaGeneratorSingleton hasn't been created!")
        return cls._instance

    def __init__(
        self,
        delta_generator_cls: type[DeltaGenerator],
        status_container_cls: type[StatusContainer],
        dialog_container_cls: type[Dialog],
        expander_container_cls: type[ExpanderContainer],
        tab_container_cls: type[TabContainer],
        popover_container_cls: type[PopoverContainer],
    ) -> None:
        """Registers and initializes all delta-generator classes.

        Parameters
        ----------
        delta_generator_cls : type[DeltaGenerator]
            The main DeltaGenerator class.
        status_container_cls : type[StatusContainer]
            The delta-generator class that is used as return value for `st.status`.
        dialog_container_cls : type[Dialog]
            The delta-generator class is used as return value for `st.dialog`.
        expander_container_cls : type[ExpanderContainer]
            The delta-generator class used as return value for `st.expander`.
        tab_container_cls : type[TabContainer]
            The delta-generator class used as return value for individual tabs
            in `st.tabs`.
        popover_container_cls : type[PopoverContainer]
            The delta-generator class used as return value for `st.popover`.

        Raises
        ------
        RuntimeError
            If the DeltaGeneratorSingleton instance already exists.
        """
        if DeltaGeneratorSingleton._instance is not None:
            raise RuntimeError("DeltaGeneratorSingleton instance already exists!")
        DeltaGeneratorSingleton._instance = self

        self._main_dg = delta_generator_cls(root_container=_RootContainer.MAIN)
        self._sidebar_dg = delta_generator_cls(
            root_container=_RootContainer.SIDEBAR, parent=self._main_dg
        )
        self._event_dg = delta_generator_cls(
            root_container=_RootContainer.EVENT, parent=self._main_dg
        )
        self._bottom_dg = delta_generator_cls(
            root_container=_RootContainer.BOTTOM, parent=self._main_dg
        )
        self._status_container_cls = status_container_cls
        self._dialog_container_cls = dialog_container_cls
        self._expander_container_cls = expander_container_cls
        self._tab_container_cls = tab_container_cls
        self._popover_container_cls = popover_container_cls

    @property
    def main_dg(self) -> DeltaGenerator:
        return self._main_dg

    @property
    def sidebar_dg(self) -> DeltaGenerator:
        return self._sidebar_dg

    @property
    def event_dg(self) -> DeltaGenerator:
        return self._event_dg

    @property
    def bottom_dg(self) -> DeltaGenerator:
        return self._bottom_dg

    @property
    def status_container_cls(
        self,
    ) -> type[StatusContainer]:
        """Stub for StatusContainer. Since StatusContainer inherits from DeltaGenerator,
        this is used to avoid circular imports.
        """
        return self._status_container_cls

    @property
    def dialog_container_cls(self) -> type[Dialog]:
        """Stub for Dialog. Since Dialog inherits from DeltaGenerator,
        this is used to avoid circular imports.
        """
        return self._dialog_container_cls

    @property
    def expander_container_cls(self) -> type[ExpanderContainer]:
        """Stub for ExpanderContainer. Since ExpanderContainer inherits from
        DeltaGenerator, this is used to avoid circular imports.
        """
        return self._expander_container_cls

    @property
    def tab_container_cls(self) -> type[TabContainer]:
        """Stub for TabContainer. Since TabContainer inherits from
        DeltaGenerator, this is used to avoid circular imports.
        """
        return self._tab_container_cls

    @property
    def popover_container_cls(self) -> type[PopoverContainer]:
        """Stub for PopoverContainer. Since PopoverContainer inherits from
        DeltaGenerator, this is used to avoid circular imports.
        """
        return self._popover_container_cls


def get_dg_singleton_instance() -> DeltaGeneratorSingleton:
    """Return the DeltaGeneratorSingleton instance. Raise an Error if the
    DeltaGeneratorSingleton hasn't been created yet.
    """
    return DeltaGeneratorSingleton.instance()


_T = TypeVar("_T")


class ContextVarWithLazyDefault(Generic[_T]):
    """The dg_stack tracks the currently active DeltaGenerator, and is pushed to when
    a DeltaGenerator is entered via a `with` block. This is implemented as a ContextVar
    so that different threads or async tasks can have their own stacks.

    We have a wrapper around it because ContextVar default cannot be a function, but
    the default dg (main_dg) might not exist yet when this module is imported.
    """

    def __init__(self, name: str, *, default: Callable[[], _T]) -> None:
        self._name = name
        self._default = default
        self._context_var: ContextVar[_T] | None = None

    def _ensure_context_var(self) -> ContextVar[_T]:
        if self._context_var is None:
            self._context_var = ContextVar(self._name, default=self._default())  # noqa: B039
        return self._context_var

    def get(self) -> _T:
        return self._ensure_context_var().get()

    def set(self, value: _T) -> Token[_T]:
        return self._ensure_context_var().set(value)

    def reset(self, token: Token[_T]) -> None:
        self._ensure_context_var().reset(token)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ContextVarWithLazyDefault):
            return NotImplemented
        # Two wrappers are equal only if they're the same object
        return self is other

    def __hash__(self) -> int:
        return self._ensure_context_var().__hash__()


# we don't use the default factory here because `main_dg` is not initialized when this
# module is imported. This is why we have our own ContextVar wrapper.
context_dg_stack: ContextVarWithLazyDefault[tuple[DeltaGenerator, ...]] = (
    ContextVarWithLazyDefault(
        "context_dg_stack", default=lambda: (get_dg_singleton_instance().main_dg,)
    )
)


def get_default_dg_stack_value() -> tuple[DeltaGenerator, ...]:
    """Get the default dg_stack value with which the dg_stack should
    be initialized and reset if needed.
    """
    instance = get_dg_singleton_instance()
    if instance._main_dg is None:
        raise RuntimeError("main_dg is not set")

    return (instance._main_dg,)


def get_last_dg_added_to_context_stack() -> DeltaGenerator | None:
    """Get the last added DeltaGenerator of the stack in the current context.

    Returns None if the stack has only one element or is empty for whatever reason.
    """
    current_stack = context_dg_stack.get()
    # If set to "> 0" and thus return the only delta generator in the stack -
    # which logically makes more sense -, some unit tests fail.
    # It looks like the reason is that they create their own main delta generator
    # but do not populate the dg_stack correctly. However, to be on the safe-side,
    # we keep the logic but leave the comment as shared knowledge for whoever will look
    # into this in the future.
    if len(current_stack) > 1:
        return current_stack[-1]
    return None



# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Accessors for the BidiComponentManager used by Custom Components v2.

This module exposes `get_bidi_component_manager`, which returns the singleton
`BidiComponentManager` registered on the active Streamlit runtime. When no
runtime is active, a local manager instance is created and returned.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from streamlit.components.v2.component_manager import BidiComponentManager


def get_bidi_component_manager() -> BidiComponentManager:
    """Return the singleton ``BidiComponentManager`` instance.

    Returns
    -------
    BidiComponentManager
        The singleton BidiComponentManager instance.


    Notes
    -----
    If the Streamlit runtime is not running, a local ``BidiComponentManager``
    is created and returned.
    """
    from streamlit.components.v2.component_manager import BidiComponentManager
    from streamlit.runtime import Runtime

    if Runtime.exists():
        return Runtime.instance().bidi_component_registry

    # Return a local manager when running without the streamlit runtime
    return BidiComponentManager()
